import os
import sys
# from .Channel import *
from .Config import *
from .Core import *
from .DataProcessing import *
from .DataStorage import *
from .Lib import *
from .Simulate import *
from .Strategy import *

sys.modules["ROOT_DIR"] = os.path.abspath(os.path.dirname(__file__))

__all__ = []
# __all__.extend(Channel.__all__)
__all__.extend(Config.__all__)
__all__.extend(Core.__all__)
__all__.extend(DataProcessing.__all__)
__all__.extend(DataStorage.__all__)
__all__.extend(Lib.__all__)
__all__.extend(Simulate.__all__)
__all__.extend(Strategy.__all__)



import os
import sys

import xml.dom.minidom
from datetime import datetime

from SapphireQuant.Core import Holiday


class HolidayHelper:
    """
    节假日
    """
    FileName = "Holiday.xml"
    _holiday_map = None
    _holidays = None
    IsLoaded = False

    @staticmethod
    def get_all_china_holidays():
        """
        加载配置
        """
        if not HolidayHelper.IsLoaded:
            try:
                config_directory = os.path.join(sys.modules['ROOT_DIR'], 'Profiles')
                path = os.path.join(config_directory, HolidayHelper.FileName)
                if os.path.exists(path):
                    pass
                else:
                    print("未找到配置文件:" + path)

                HolidayHelper._holidays = []
                HolidayHelper._holiday_map = {}
                _root = xml.dom.minidom.parse(path).documentElement
                holiday_nodes = _root.getElementsByTagName('Holiday')
                for holiday_node in holiday_nodes:
                    holiday = Holiday()
                    holiday.name = holiday_node.getAttribute('name')
                    holiday.pre_holiday_trading_date = datetime.strptime(holiday_node.getAttribute('pre_holiday_trading_date'), '%Y%m%d')
                    holiday.begin_date = datetime.strptime(holiday_node.getAttribute('begin_date'), '%Y%m%d')
                    holiday.end_date = datetime.strptime(holiday_node.getAttribute('end_date'), '%Y%m%d')
                    holiday.post_holiday_trading_date = datetime.strptime(holiday_node.getAttribute('post_holiday_trading_date'), '%Y%m%d')
                    HolidayHelper._holidays.append(holiday)
                    HolidayHelper._holiday_map[holiday.post_holiday_trading_date] = holiday

                HolidayHelper.IsLoaded = True
            except Exception as e:
                print(str(e))
        pass

    @staticmethod
    def holidays():
        """
        节假日
        :return:
        """
        if HolidayHelper._holidays is None:
            HolidayHelper.get_all_china_holidays()
        return HolidayHelper._holidays

    @staticmethod
    def is_post_holiday_trading_date(date):
        """
        是否为节假日后第一个交易日
        :return:
        """
        if HolidayHelper._holidays is None:
            HolidayHelper.get_all_china_holidays()

        if date in HolidayHelper._holiday_map.keys():
            return True
        return False

# date = datetime(2020,1,2)
# a = HolidayHelper.is_post_holiday_trading_date(date)
# print(a)



from SapphireQuant.Core.Entity import Entity


class FutureAsset(Entity):
    """
    期货资金明细
    """

    def __init__(self):
        super().__init__()
        self.cash = 0
        self.pre_fund_balance = 0
        self.fund_balance = 0
        self.available_cash = 0
        self.frozen_cash = 0
        self.market_value = 0
        self.position_profit = 0
        self.float_profit = 0
        self.close_profit = 0
        self.acc_close_profit = 0
        self.fee = 0
        self.current_margin = 0
        self.frozen_margin = 0
        self.margin = 0
        self.withdrawal = 0  # 出入金
        self.delivery_fee = 0  # 交割手续费

    def clone(self):
        """

        :return:
        """
        asset = FutureAsset()
        asset.cash = self.cash
        asset.pre_fund_balance = self.pre_fund_balance
        asset.fund_balance = self.fund_balance
        asset.available_cash = self.available_cash
        asset.frozen_cash = self.frozen_cash
        asset.market_value = self.market_value
        asset.position_profit = self.position_profit
        asset.float_profit = self.float_profit
        asset.close_profit = self.close_profit
        asset.acc_close_profit = self.acc_close_profit
        asset.fee = self.fee
        asset.current_margin = self.current_margin
        asset.frozen_margin = self.frozen_margin
        asset.margin = self.margin
        asset.withdrawal = self.withdrawal
        asset.delivery_fee = self.delivery_fee
        return asset

    def to_string(self):
        """
        string输出
        """
        return '日期:{0},昨日结存:{1},动态权益:{2},平仓盈亏:{3},持仓盈亏:{4},净利润:{5},可用资金:{6},持仓保证金:{7},手续费:{8},出入金:{9},交割手续费:{10}'.format(
            self.trading_date.strftime('%Y/%m/%d'), self.pre_fund_balance, self.fund_balance, self.close_profit, self.position_profit,
            self.close_profit + self.position_profit, self.available_cash, self.margin, self.fee, self.withdrawal, self.delivery_fee
        )



import datetime


from SapphireQuant.Core import Tick, Quote

from SapphireQuant.Core.Futures import FutureOrder, FutureTrade, FuturePosition, FutureAsset, FuturePositionDetail

from SapphireQuant.Core.Enum import EnumBuySell, EnumPositionDirection, EnumOpenClose, EnumHedgeFlag, EnumOrderStatus, EnumOrderSubmitStatus, EnumMarket, EnumPositionType, EnumProductClass
from SapphireQuant.Core.Enum.EnumOrderPriceType import EnumOrderPriceType
from SapphireQuant.Core.Enum.EnumOrderRejectReason import EnumOrderRejectReason
from SapphireQuant.Core.Enum.EnumOrderType import EnumOrderType
from SapphireQuant.Channel.CtpChannel import thosttraderapi as api


class CtpConvert:
    """
    Ctp转换
    """
    @staticmethod
    def convert_product_class(ctp_product_class):
        """
        转换柜台的ProductClass
        :param ctp_product_class:
        """
        product_class = None
        if ctp_product_class == api.THOST_FTDC_PC_Futures:
            product_class = EnumProductClass.future
        elif ctp_product_class == api.THOST_FTDC_PC_Options:
            product_class = EnumProductClass.option
        elif ctp_product_class == api.THOST_FTDC_PC_Combination:
            product_class = EnumProductClass.combination
        elif ctp_product_class == api.THOST_FTDC_PC_Spot:
            product_class = EnumProductClass.spot
        elif ctp_product_class == api.THOST_FTDC_PC_EFP:
            product_class = EnumProductClass.efp
        elif ctp_product_class == api.THOST_FTDC_PC_SpotOption:
            product_class = EnumProductClass.spot_option
        return product_class

    @staticmethod
    def convert_tick(ctp_tick):
        """

        :param ctp_tick:
        """
        tick = Tick()
        tick.market = EnumMarket.期货
        tick.trading_date = datetime.datetime.strptime(ctp_tick.TradingDay, '%Y%m%d')
        tick.instrument_id = ctp_tick.InstrumentID
        tick.exchange_id = ctp_tick.ExchangeID
        #  ctp_tick.ExchangeInstID
        tick.last_price = ctp_tick.LastPrice
        tick.pre_settlement_price = ctp_tick.PreSettlementPrice
        tick.pre_close_price = ctp_tick.PreClosePrice
        tick.pre_open_interest = ctp_tick.PreOpenInterest
        tick.open_price = ctp_tick.OpenPrice
        tick.high_price = ctp_tick.HighestPrice
        tick.low_price = ctp_tick.LowestPrice
        tick.volume = ctp_tick.Volume
        tick.turnover = ctp_tick.Turnover
        tick.open_interest = ctp_tick.OpenInterest
        #  ctp_tick.ClosePrice,
        #  ctp_tick.SettlementPrice,
        tick.up_limit = ctp_tick.UpperLimitPrice
        tick.drop_limit = ctp_tick.LowerLimitPrice
        #  ctp_tick.PreDelta
        #  ctp_tick.CurrDlta
        tick.date_time = datetime.datetime.strptime('{0} {1}'.format(ctp_tick.TradingDay, ctp_tick.UpdateTime), '%Y%m%d %H:%M:%S')
        #  ctp_tick.UpdateMillisec
        tick.bid_price1 = ctp_tick.BidPrice1
        tick.bid_volume1 = ctp_tick.BidVolume1
        tick.ask_price1 = ctp_tick.AskPrice1
        tick.ask_volume1 = ctp_tick.AskVolume1
        tick.quote = Quote()
        tick.quote.bid_price1 = ctp_tick.BidPrice1
        tick.quote.bid_volume1 = ctp_tick.BidVolume1
        tick.quote.ask_price1 = ctp_tick.AskPrice1
        tick.quote.ask_volume1 = ctp_tick.AskVolume1

        tick.quote.bid_price2 = ctp_tick.BidPrice2
        tick.quote.bid_volume2 = ctp_tick.BidVolume2
        tick.quote.ask_price2 = ctp_tick.AskPrice2
        tick.quote.ask_volume2 = ctp_tick.AskVolume2

        tick.quote.bid_price3 = ctp_tick.BidPrice3
        tick.quote.bid_volume3 = ctp_tick.BidVolume3
        tick.quote.ask_price3 = ctp_tick.AskPrice3
        tick.quote.ask_volume3 = ctp_tick.AskVolume3

        tick.quote.bid_price4 = ctp_tick.BidPrice4
        tick.quote.bid_volume4 = ctp_tick.BidVolume4
        tick.quote.ask_price4 = ctp_tick.AskPrice4
        tick.quote.ask_volume4 = ctp_tick.AskVolume4

        tick.quote.bid_price5 = ctp_tick.BidPrice5
        tick.quote.bid_volume5 = ctp_tick.BidVolume5
        tick.quote.ask_price5 = ctp_tick.AskPrice5
        tick.quote.ask_volume5 = ctp_tick.AskVolume5
        #  ctp_tick.AveragePrice,
        #  ctp_tick.ActionDay,
        tick.date_time = tick.date_time + datetime.timedelta(milliseconds=int(ctp_tick.UpdateMillisec))
        tick.local_time = datetime.datetime.now()
        return tick

    @staticmethod
    def convert_direction(ctp_direction):
        """
        转换柜台的Direction
        :param ctp_direction:
        """
        direction = None
        if ctp_direction == api.THOST_FTDC_D_Buy:
            direction = EnumBuySell.买入
        elif ctp_direction == api.THOST_FTDC_D_Sell:
            direction = EnumBuySell.卖出
        return direction

    @staticmethod
    def convert_position_direction(ctp_position_direction):
        """
        转换柜台的Direction
        :param ctp_position_direction:
        """
        position_direction = None
        if ctp_position_direction == api.THOST_FTDC_PD_Long:
            position_direction = EnumPositionDirection.多头
        elif ctp_position_direction == api.THOST_FTDC_PD_Short:
            position_direction = EnumPositionDirection.空头
        elif ctp_position_direction == api.THOST_FTDC_PD_Net:
            position_direction = EnumPositionDirection.净持仓
        return position_direction

    @staticmethod
    def convert_open_close(ctp_open_close):
        """
        转换柜台的OpenClose
        :param ctp_open_close:
        """
        open_close = None
        if ctp_open_close == api.THOST_FTDC_OF_Open:
            open_close = EnumOpenClose.开仓
        elif ctp_open_close == api.THOST_FTDC_OF_Close:
            open_close = EnumOpenClose.平仓
        elif ctp_open_close == api.THOST_FTDC_OF_ForceClose:
            open_close = EnumOpenClose.强平
        elif ctp_open_close == api.THOST_FTDC_OF_CloseToday:
            open_close = EnumOpenClose.平今
        elif ctp_open_close == api.THOST_FTDC_OF_CloseYesterday:
            open_close = EnumOpenClose.平昨
        elif ctp_open_close == api.THOST_FTDC_OF_ForceOff:
            open_close = EnumOpenClose.强减
        elif ctp_open_close == api.THOST_FTDC_OF_LocalForceClose:
            open_close = EnumOpenClose.本地强平
        return open_close

    @staticmethod
    def convert_hedge_flag(ctp_hedge_flag):
        """
        转换柜台的OpenClose
        :param ctp_hedge_flag:
        """
        hedge_flag = None
        if ctp_hedge_flag == api.THOST_FTDC_BHF_Speculation:
            hedge_flag = EnumHedgeFlag.投机
        elif ctp_hedge_flag == api.THOST_FTDC_BHF_Arbitrage:
            hedge_flag = EnumHedgeFlag.套利
        elif ctp_hedge_flag == api.THOST_FTDC_BHF_Hedge:
            hedge_flag = EnumHedgeFlag.套保
        return hedge_flag

    @staticmethod
    def convert_order_type(ctp_order_type):
        """
        转换柜台的OpenClose
        :param ctp_order_type:
        """
        order_type = None
        if ctp_order_type == api.THOST_FTDC_ORDT_Normal:
            order_type = EnumOrderType.normal
        elif ctp_order_type == api.THOST_FTDC_ORDT_DeriveFromQuote:
            order_type = EnumOrderType.derive_from_quote
        elif ctp_order_type == api.THOST_FTDC_ORDT_DeriveFromCombination:
            order_type = EnumOrderType.derivce_from_combination
        elif ctp_order_type == api.THOST_FTDC_ORDT_Combination:
            order_type = EnumOrderType.combination
        elif ctp_order_type == api.THOST_FTDC_ORDT_ConditionalOrder:
            order_type = EnumOrderType.conditional_order
        elif ctp_order_type == api.THOST_FTDC_ORDT_Swap:
            order_type = EnumOrderType.swap
        elif ctp_order_type == api.THOST_FTDC_ORDT_DeriveFromBlockTrade:
            order_type = EnumOrderType.derive_from_block_trade
        elif ctp_order_type == api.THOST_FTDC_ORDT_DeriveFromEFPTrade:
            order_type = EnumOrderType.derive_from_EFP_trade
        return order_type

    @staticmethod
    def convert_order_reject_reason(error_msg):
        """
        转换柜台的OpenClose
        :param error_msg:
        """
        order_reject_reason = EnumOrderRejectReason.未知
        if "每秒请求超许可数" in error_msg:
            order_reject_reason = EnumOrderRejectReason.每秒请求超许可数
        if "价格非最小单位倍数" in error_msg:
            order_reject_reason = EnumOrderRejectReason.价格非最小单位倍数
        if "价格高于涨停板" in error_msg:
            order_reject_reason = EnumOrderRejectReason.价格高于涨停板
        if "价格低于跌停板" in error_msg:
            order_reject_reason = EnumOrderRejectReason.价格低于跌停板
        if "资金不足" in error_msg:
            order_reject_reason = EnumOrderRejectReason.资金不足
        if "平今仓位不足" in error_msg:
            order_reject_reason = EnumOrderRejectReason.平今仓位不足
        if "平昨仓位不足" in error_msg:
            order_reject_reason = EnumOrderRejectReason.平昨仓位不足
        if "不支持报单类型" in error_msg:
            order_reject_reason = EnumOrderRejectReason.不支持报单类型
        if "不支持套利类型报单" in error_msg:
            order_reject_reason = EnumOrderRejectReason.不支持套利类型报单
        if "保值额度不足" in error_msg:
            order_reject_reason = EnumOrderRejectReason.保值额度不足
        if "网络连接失败" in error_msg:
            order_reject_reason = EnumOrderRejectReason.网络连接失败
        if "Qi暂停开仓" in error_msg:
            order_reject_reason = EnumOrderRejectReason.Qi暂停开仓
        if "延迟开仓被Qi拒绝" in error_msg:
            order_reject_reason = EnumOrderRejectReason.延迟开仓被Qi拒绝
        if "报单数量错误" in error_msg:
            order_reject_reason = EnumOrderRejectReason.报单数量错误
        if "提交委托失败" in error_msg:
            order_reject_reason = EnumOrderRejectReason.提交委托失败
        return order_reject_reason

    @staticmethod
    def convert_order_price_type(ctp_order_price_type):
        """
        转换柜台的OrderPriceType
        :param ctp_order_price_type:
        """
        order_price_type = None
        if ctp_order_price_type == api.THOST_FTDC_OPT_AnyPrice:
            order_price_type = EnumOrderPriceType.市价
        elif ctp_order_price_type == api.THOST_FTDC_OPT_LimitPrice:
            order_price_type = EnumOrderPriceType.限价
        elif ctp_order_price_type == api.THOST_FTDC_OPT_BestPrice:
            order_price_type = EnumOrderPriceType.最优价
        elif ctp_order_price_type == api.THOST_FTDC_OPT_LastPrice:
            order_price_type = EnumOrderPriceType.最新价
        elif ctp_order_price_type == api.THOST_FTDC_OPT_LastPricePlusOneTicks:
            order_price_type = EnumOrderPriceType.最新价浮动上浮1个ticks
        elif ctp_order_price_type == api.THOST_FTDC_OPT_LastPricePlusTwoTicks:
            order_price_type = EnumOrderPriceType.最新价浮动上浮2个ticks
        elif ctp_order_price_type == api.THOST_FTDC_OPT_LastPricePlusThreeTicks:
            order_price_type = EnumOrderPriceType.最新价浮动上浮3个ticks
        elif ctp_order_price_type == api.THOST_FTDC_OPT_AskPrice1:
            order_price_type = EnumOrderPriceType.卖一价
        elif ctp_order_price_type == api.THOST_FTDC_OPT_AskPrice1PlusOneTicks:
            order_price_type = EnumOrderPriceType.卖一价浮动上浮1个ticks
        elif ctp_order_price_type == api.THOST_FTDC_OPT_AskPrice1PlusTwoTicks:
            order_price_type = EnumOrderPriceType.卖一价浮动上浮2个ticks
        elif ctp_order_price_type == api.THOST_FTDC_OPT_AskPrice1PlusThreeTicks:
            order_price_type = EnumOrderPriceType.卖一价浮动上浮3个ticks
        elif ctp_order_price_type == api.THOST_FTDC_OPT_BidPrice1:
            order_price_type = EnumOrderPriceType.买一价
        elif ctp_order_price_type == api.THOST_FTDC_OPT_BidPrice1PlusOneTicks:
            order_price_type = EnumOrderPriceType.买一价浮动上浮1个ticks
        elif ctp_order_price_type == api.THOST_FTDC_OPT_BidPrice1PlusTwoTicks:
            order_price_type = EnumOrderPriceType.买一价浮动上浮2个ticks
        elif ctp_order_price_type == api.THOST_FTDC_OPT_BidPrice1PlusThreeTicks:
            order_price_type = EnumOrderPriceType.买一价浮动上浮3个ticks
        return order_price_type

    @staticmethod
    def convert_order_status(ctp_order_status):
        """
        转换柜台的OrderPriceType
        :param ctp_order_status:
        """
        order_status = None
        if ctp_order_status == api.THOST_FTDC_OST_AllTraded:
            order_status = EnumOrderStatus.全部成交
        elif ctp_order_status == api.THOST_FTDC_OST_PartTradedQueueing:
            order_status = EnumOrderStatus.部分成交
        elif ctp_order_status == api.THOST_FTDC_OST_PartTradedNotQueueing:
            order_status = EnumOrderStatus.正撤
        elif ctp_order_status == api.THOST_FTDC_OST_NoTradeQueueing:
            order_status = EnumOrderStatus.已报
        elif ctp_order_status == api.THOST_FTDC_OST_NoTradeNotQueueing:
            order_status = EnumOrderStatus.正撤
        elif ctp_order_status == api.THOST_FTDC_OST_Canceled:
            order_status = EnumOrderStatus.已撤
        elif ctp_order_status == api.THOST_FTDC_OST_Unknown:
            order_status = EnumOrderStatus.未知
        elif ctp_order_status == api.THOST_FTDC_OST_NotTouched:
            order_status = EnumOrderStatus.未知
        elif ctp_order_status == api.THOST_FTDC_OST_Touched:
            order_status = EnumOrderStatus.未知
        return order_status

    @staticmethod
    def convert_order_submit_status(ctp_order_status):
        """
        转换柜台的OrderPriceType
        :param ctp_order_status:
        """
        order_submit_status = None
        if ctp_order_status == api.THOST_FTDC_OSS_InsertSubmitted:
            order_submit_status = EnumOrderSubmitStatus.insert_submitted
        elif ctp_order_status == api.THOST_FTDC_OSS_CancelSubmitted:
            order_submit_status = EnumOrderSubmitStatus.cancel_submitted
        elif ctp_order_status == api.THOST_FTDC_OSS_ModifySubmitted:
            order_submit_status = EnumOrderSubmitStatus.modify_submitted
        elif ctp_order_status == api.THOST_FTDC_OSS_Accepted:
            order_submit_status = EnumOrderSubmitStatus.accepted
        elif ctp_order_status == api.THOST_FTDC_OSS_InsertRejected:
            order_submit_status = EnumOrderSubmitStatus.insert_rejected
        elif ctp_order_status == api.THOST_FTDC_OSS_CancelRejected:
            order_submit_status = EnumOrderSubmitStatus.cancel_rejected
        elif ctp_order_status == api.THOST_FTDC_OSS_ModifyRejected:
            order_submit_status = EnumOrderSubmitStatus.modify_rejected
        return order_submit_status

    @staticmethod
    def convert_asset(ctp_asset):
        """
        转换柜台的Account到本地
        :param ctp_asset:
        :return:
        """
        asset = FutureAsset()
        # _asset.BrokerID = ctp_asset.BrokerID
        # _asset.AccountID = ctp_asset.AccountID
        # _asset.PreMortgage = ctp_asset.PreMortgage
        # _asset.PreCredit = ctp_asset.PreCredit
        # _asset.PreDeposit = ctp_asset.PreDeposit
        asset.pre_fund_balance = ctp_asset.PreBalance
        # _asset.PreMargin = ctp_asset.PreMargin
        # _asset.InterestBase = ctp_asset.InterestBase
        # _asset.Interest = ctp_asset.Interest
        # _asset.Deposit = ctp_asset.Deposit
        # _asset.Withdraw = ctp_asset.Withdraw
        asset.frozen_margin = ctp_asset.FrozenMargin
        asset.frozen_cash = ctp_asset.FrozenCash
        # _asset.FrozenCommission = ctp_asset.FrozenCommission
        asset.current_margin = ctp_asset.CurrMargin
        # _asset.CashIn = ctp_asset.CashIn
        # _asset.Commission = ctp_asset.Commission
        asset.close_profit = ctp_asset.CloseProfit
        asset.position_profit = ctp_asset.PositionProfit
        asset.fund_balance = ctp_asset.Balance
        asset.available_cash = ctp_asset.Available
        # _asset.WithdrawQuota = ctp_asset.WithdrawQuota
        # _asset.Reserve = ctp_asset.Reserve
        asset.trading_date = datetime.datetime.strptime(ctp_asset.TradingDay, '%Y%m%d')
        # _asset.SettlementID = ctp_asset.SettlementID
        # _asset.Credit = ctp_asset.Credit
        # _asset.Mortgage = ctp_asset.Mortgage
        # _asset.ExchangeMargin = ctp_asset.ExchangeMargin
        # _asset.DeliveryMargin = ctp_asset.DeliveryMargin
        # _asset.ExchangeDeliveryMargin = ctp_asset.ExchangeDeliveryMargin
        # _asset.ReserveBalance = ctp_asset.ReserveBalance
        return asset

    @staticmethod
    def convert_order(ctp_order):
        """
        转换柜台Order
        :param ctp_order:
        :return:
        """
        try:
            order_ref = int(ctp_order.OrderRef)
            order = FutureOrder()
            order.broker_id = ctp_order.BrokerID
            order.investor_id = ctp_order.InvestorID
            order.instrument_id = ctp_order.InstrumentID
            order.local_id = CtpConvert.get_ctp_unique_local_id(ctp_order.FrontID, ctp_order.SessionID, order_ref, ctp_order.InstrumentID)
            order.user_id = ctp_order.UserID,
            # ctp_order.PriceType,
            order.direction = CtpConvert.convert_direction(ctp_order.Direction)
            # OpenClose = ctp_order.CombOffsetFlag,
            # HedgeFlag = ctp_order.CombHedgeFlag,
            order.order_price = ctp_order.LimitPrice
            order.order_volume = ctp_order.VolumeTotalOriginal
            # ctp_order.TimeCondition,
            # ctp_order.GTDDate,
            # ctp_order.VolumeCondition,
            # ctp_order.MinVolume,
            # ctp_order.ContingentCondition,
            # ctp_order.StopPrice,
            # ctp_order.ForceCloseReason,
            # ctp_order.IsAutoSuspend,
            # ctp_order.BusinessUnit,
            # ctp_order.RequestID,
            # ctp_order.OrderLocalID,
            order.exchange_id = ctp_order.ExchangeID
            # ctp_order.ParticipantID,
            # ctp_order.ClientID,
            # ctp_order.ExchangeInstID,
            # ctp_order.TraderID,
            # ctp_order.InstallID,
            order.order_submit_status = CtpConvert.convert_order_submit_status(ctp_order.OrderSubmitStatus)
            # ctp_order.NotifySequence,
            order.trading_date = datetime.datetime.strptime(ctp_order.TradingDay, '%Y%m%d')
            # ctp_order.SettlementID,
            order.order_sys_id = ctp_order.OrderSysID
            # ctp_order.OrderSource,
            order.order_status = CtpConvert.convert_order_status(ctp_order.OrderStatus)
            order.order_type = CtpConvert.convert_order_type(ctp_order.OrderType)
            order.volume_traded = ctp_order.VolumeTraded
            # ctp_order.VolumeTotal,# 表示这笔委托还没有成交的数量
            # ctp_order.ActiveTime,
            # ctp_order.SuspendTime,
            # ctp_order.UpdateTime
            # ctp_order.CancelTime
            order.order_time = datetime.datetime.strptime('{0} {1}'.format(ctp_order.InsertDate, ctp_order.InsertTime), '%Y%m%d %H:%M:%S')
            if ctp_order.CancelTime.strip() == '':
                order.cancel_time = order.order_time
            else:
                order.cancel_time = datetime.datetime.strptime('{0} {1}'.format(ctp_order.InsertDate, ctp_order.CancelTime), '%Y%m%d %H:%M:%S')
            # ctp_order.ActiveTraderID,
            # ctp_order.ClearingPartID,
            # ctp_order.SequenceNo,
            # ctp_order.FrontID,
            # ctp_order.SessionID,
            # ctp_order.UserProductInfo,
            order.status_msg = ctp_order.StatusMsg
            # ctp_order.UserForceClose,
            # ctp_order.ActiveUserID,
            # ctp_order.BrokerOrderSeq,
            # ctp_order.RelativeOrderSysID,
            # ctp_order.ZCETotalTradedVolume,
            # ctp_order.IsSwapOrder
            return order
        except Exception as e:
            print(str(e))
        return None

    @staticmethod
    def convert_input_order(ctp_input_order, rsp_info, front_id, session_id):
        """

        :param ctp_input_order:
        :param rsp_info:
        :param front_id:
        :param session_id:
        :return: 
        """
        order_ref = int(ctp_input_order.OrderRef)
        order = FutureOrder()
        order.broker_id = ctp_input_order.BrokerID
        order.investor_id = ctp_input_order.InvestorID
        order.instrument_id = ctp_input_order.InstrumentID
        order.local_id = CtpConvert.get_ctp_unique_local_id(front_id, session_id, order_ref, ctp_input_order.InstrumentID)
        order.user_id = ctp_input_order.UserID
        order.price_type = CtpConvert.convert_order_price_type(ctp_input_order.OrderPriceType)
        order.direction = CtpConvert.convert_direction(ctp_input_order.Direction)
        #  ctp_input_order.CombOffsetFlag_0
        #  ctp_input_order.CombOffsetFlag_1
        #  ctp_input_order.CombOffsetFlag_2
        #  ctp_input_order.CombOffsetFlag_3
        #  ctp_input_order.CombOffsetFlag_4
        #  ctp_input_order.CombHedgeFlag_0
        #  ctp_input_order.CombHedgeFlag_1
        #  ctp_input_order.CombHedgeFlag_2
        #  ctp_input_order.CombHedgeFlag_3
        #  ctp_input_order.CombHedgeFlag_4
        order.order_price = ctp_input_order.LimitPrice
        order.order_volume = ctp_input_order.VolumeTotalOriginal
        #  ctp_input_order.TimeCondition
        #  ctp_input_order.GTDDate
        #  ctp_input_order.VolumeCondition
        #  ctp_input_order.MinVolume
        #  ctp_input_order.ContingentCondition
        #  ctp_input_order.StopPrice
        #  ctp_input_order.ForceCloseReason
        #  ctp_input_order.IsAutoSuspend
        #  ctp_input_order.BusinessUnit
        #  ctp_input_order.RequestID
        #  ctp_input_order.UserForceClose
        #  ctp_input_order.IsSwapOrder
        order.status_msg = rsp_info.ErrorMsg
        order.order_reject_reason = CtpConvert.convert_order_reject_reason(rsp_info.ErrorMsg)
        order.order_status = EnumOrderStatus.废单

        return order

    @staticmethod
    def convert_order_action(ctp_order_action, rsp_info):
        """

        :param ctp_order_action:
        :param rsp_info:
        :return:
        """
        order_ref = int(ctp_order_action.OrderRef)
        order = FutureOrder()
        order.broker_id = ctp_order_action.BrokerID
        order.investor_id = ctp_order_action.InvestorID
        order.instrument_id = ctp_order_action.InstrumentID
        order.local_id = CtpConvert.get_ctp_unique_local_id(ctp_order_action.FrontID, ctp_order_action.SessionID, order_ref, ctp_order_action.InstrumentID)
        order.user_id = ctp_order_action.UserID
        order.status_msg = rsp_info.ErrorMsg
        order.exchange_id = ctp_order_action.ExchangeID
        order.order_price = ctp_order_action.LimitPrice
        order.order_sys_id = ctp_order_action.OrderSysID

        return order

    @staticmethod
    def convert_input_order_action(order_action, rsp_info):
        """

        :param order_action:
        :param rsp_info:
        :return:
        """
        order_ref = int(order_action.OrderRef)
        order = FutureOrder()
        order.BrokerId = order_action.BrokerID
        order.InvestorId = order_action.InvestorID
        order.InstrumentId = order_action.InstrumentID
        order.LocalId = CtpConvert.get_ctp_unique_local_id(order_action.FrontID, order_action.SessionID, order_ref, order_action.InstrumentID)
        order.UserId = order_action.UserID
        order.StatusMsg = rsp_info.ErrorMsg
        order.ExchangeId = order_action.ExchangeID
        order.OrderPrice = order_action.LimitPrice
        order.OrderSysId = order_action.OrderSysID

        return order

    @staticmethod
    def convert_trade(ctp_trade, front_id, session_id):
        """

        :param ctp_trade:
        :param front_id:
        :param session_id:
        :return:
        """
        try:
            order_ref = int(ctp_trade.OrderRef)
            trade = FutureTrade()
            trade.broker_id = ctp_trade.BrokerID
            trade.investor_id = ctp_trade.InvestorID
            trade.instrument_id = ctp_trade.InstrumentID
            trade.local_id = CtpConvert.get_ctp_unique_local_id(front_id, session_id, order_ref, ctp_trade.InstrumentID)
            trade.user_id = ctp_trade.UserID
            trade.exchange_id = ctp_trade.ExchangeID
            trade.trade_sys_id = ctp_trade.TradeID
            trade.direction = CtpConvert.convert_direction(ctp_trade.Direction)
            trade.order_sys_id = ctp_trade.OrderSysID
            #  ctp_trade.ParticipantID
            #  ctp_trade.ClientID
            #  ctp_trade.TradingRole
            #  ctp_trade.ExchangeInstID
            trade.open_close = CtpConvert.convert_open_close(ctp_trade.OffsetFlag)
            trade.hedge_flag = CtpConvert.convert_hedge_flag(ctp_trade.HedgeFlag)
            trade.trade_price = ctp_trade.Price
            trade.trade_volume = ctp_trade.Volume
            trade.trade_time = datetime.datetime.strptime('{0} {1}'.format(ctp_trade.TradeDate, ctp_trade.TradeTime), '%Y%m%d %H:%M:%S')
            #  ctp_trade.TradeType
            #  ctp_trade.PriceSource
            #  ctp_trade.TraderID
            #  ctp_trade.OrderLocalID
            #  ctp_trade.ClearingPartID
            #  ctp_trade.BusinessUnit
            #  ctp_trade.SequenceNo
            trade.trading_date = datetime.datetime.strptime(ctp_trade.TradingDay, '%Y%m%d')
            #  ctp_trade.SettlementID
            #  ctp_trade.BrokerOrderSeq
            #  ctp_trade.TradeSource
            return trade
        except Exception as e:
            print(str(e))
        return None

    @staticmethod
    def convert_position(ctp_investor_position):
        """

        :param ctp_investor_position:
        :return:
        """
        position = FuturePosition()
        position.instrument_id = ctp_investor_position.InstrumentID
        #  ctp_investor_position.BrokerID
        #  ctp_investor_position.InvestorID
        position.direction = CtpConvert.convert_position_direction(ctp_investor_position.PosiDirection)
        position.hedge_flag = CtpConvert.convert_hedge_flag(ctp_investor_position.HedgeFlag)
        #  ctp_investor_position.PositionDate
        position.yd_volume = ctp_investor_position.YdPosition
        position.volume = ctp_investor_position.Position
        #  ctp_investor_position.LongFrozen
        #  ctp_investor_position.ShortFrozen
        #  ctp_investor_position.LongFrozenAmount
        #  ctp_investor_position.ShortFrozenAmount
        position.open_volume = ctp_investor_position.OpenVolume
        position.close_volume = ctp_investor_position.CloseVolume
        #  ctp_investor_position.OpenAmount
        #  ctp_investor_position.CloseAmount
        #  ctp_investor_position.PositionCost
        #  ctp_investor_position.PreMargin
        position.margin = ctp_investor_position.UseMargin
        #  ctp_investor_position.FrozenMargin
        #  ctp_investor_position.FrozenCash
        #  ctp_investor_position.FrozenCommission
        #  ctp_investor_position.CashIn
        #  ctp_investor_position.Commission
        position.close_profit = ctp_investor_position.CloseProfit
        position.position_profit = ctp_investor_position.PositionProfit
        #  ctp_investor_position.PreSettlementPrice
        #  ctp_investor_position.SettlementPrice
        position.trading_date = datetime.datetime.strptime(ctp_investor_position.TradingDay, '%Y%m%d')
        #  ctp_investor_position.SettlementID
        position.open_cost = 0  # ctp_investor_position.PositionCost / ctp_investor_position.Position * Multipler 持仓成本/(持仓数量*合约乘数) 需要接入合约明细计算
        position.market_value = 0  # 需要接入行情获取最新价结算
        #  ctp_investor_position.ExchangeMargin
        #  ctp_investor_position.CombPosition
        #  ctp_investor_position.CombLongFrozen
        #  ctp_investor_position.CombShortFrozen
        #  ctp_investor_position.CloseProfitByDate
        #  ctp_investor_position.CloseProfitByTrade
        position.td_volume = ctp_investor_position.TodayPosition
        #  ctp_investor_position.MarginRateByMoney
        #  ctp_investor_position.MarginRateByVolume
        return position

    @staticmethod
    def convert_position_detail(ctp_investor_position_detail):
        """

        :param ctp_investor_position_detail:
        :return:
        """
        position = FuturePositionDetail()
        position.instrument_id = ctp_investor_position_detail.InstrumentID
        position.broker_id = ctp_investor_position_detail.BrokerID
        position.investor_id = ctp_investor_position_detail.InvestorID
        position.hedge_flag = CtpConvert.convert_hedge_flag(ctp_investor_position_detail.HedgeFlag)
        position.direction = CtpConvert.convert_direction(ctp_investor_position_detail.Direction)
        position.open_time = datetime.datetime.strptime(ctp_investor_position_detail.OpenDate, '%Y%m%d')
        position.trade_sys_id = ctp_investor_position_detail.TradeID
        position.volume = ctp_investor_position_detail.Volume
        position.open_price = ctp_investor_position_detail.OpenPrice
        position.trading_date = datetime.datetime.strptime(ctp_investor_position_detail.TradingDay, '%Y%m%d')
        #  ctp_investor_position_detail.SettlementID
        #  ctp_investor_position_detail.TradeType
        # position.open_volume = ctp_investor_position_detail.CombInstrumentID
        position.exchange_id = ctp_investor_position_detail.ExchangeID
        position.float_profit = 0  # 需要接入行情计算
        #  ctp_investor_position_detail.CloseProfitByDate
        #  ctp_investor_position_detail.CloseProfitByTrade
        #  ctp_investor_position_detail.PositionProfitByDate
        #  ctp_investor_position_detail.PositionProfitByTrade
        position.position_profit = 0  # 需要接入行情计算
        #  ctp_investor_position_detail.Margin
        #  ctp_investor_position_detail.ExchMargin
        #  ctp_investor_position_detail.MarginRateByMoney
        #  ctp_investor_position_detail.MarginRateByVolume
        #  ctp_investor_position_detail.LastSettlementPrice
        #  ctp_investor_position_detail.SettlementPrice
        #  ctp_investor_position_detail.CloseVolume
        #  ctp_investor_position_detail.CloseAmount
        #  ctp_investor_position_detail.TimeFirstVolume
        #  ctp_investor_position_detail.InvestUnitID
        if position.open_time < position.trading_date:
            position.position_type = EnumPositionType.昨仓
        else:
            position.position_type = EnumPositionType.今仓
        position.market_value = 0  # 需要接入行情计算
        return position

    @staticmethod
    def get_ctp_unique_local_id(front_id, session_id, order_ref, instrument_id):
        """

        :param front_id:
        :param session_id:
        :param order_ref:
        :param instrument_id:
        :return:
        """
        return '{0}_{1}_{2}_{3}'.format(
            front_id, session_id, abs(session_id % 100000) * 1000000, order_ref, instrument_id
        )

    @staticmethod
    def analysis_ctp_local_id(local_id):
        """

        :param local_id: 
        :return: 
        """
        try:
            print('local_id:{0}'.format(local_id))
            local_array = local_id.split('_')
            front_id = int(local_array[0])
            session_id = int(local_array[1])
            order_ref = int(local_array[2][:len(local_array[2]) - 6])
            instrument_id = local_array[3]
            print('front_id:{0}'.format(front_id))
            print('session_id:{0}'.format(session_id))
            print('order_ref:{0}'.format(order_ref))
            print('instrument_id:{0}'.format(instrument_id))
            return {
                'success': True,
                'front_id': front_id,
                'session_id': session_id,
                'order_ref': order_ref,
                'instrument_id': instrument_id,
            }
        except Exception as e:
            front_id = -1
            session_id = -1
            order_ref = -1
            instrument_id = ""
            print("analysis_ctp_local_id Failed:{0}".format(local_id))
            print(str(e))
            return {
                'success': False,
                'front_id': front_id,
                'session_id': session_id,
                'order_ref': order_ref,
                'instrument_id': instrument_id,
            }



import sys
import datetime


class CSharpUtils:
    """
    CSharp工具类
    """
    @staticmethod
    def convert_c_sharp_ticks_to_linux_ticks(value):
        """
        将CSharp Ticks转换成Linux Ticks
        :param value:
        :return:
        """
        if value == 0:
            return value
        c_sharp_ticks1970 = 621355968000000000
        linux_ticks = (value - c_sharp_ticks1970) / 10000000
        time_zone = 28800
        linux_ticks = linux_ticks - time_zone
        return linux_ticks

    @staticmethod
    def convert_c_sharp_ticks_to_py_date_time(value):
        """
        将CSharp Ticks转换成python datetime
        :param value:
        :return:
        """
        py_time = datetime.datetime.fromtimestamp(CSharpUtils.convert_c_sharp_ticks_to_linux_ticks(value))
        return py_time

    @staticmethod
    def get_c_sharp_ticks(value):
        """

        :param value:
        :return:
        """
        c_sharp_ticks1970 = 621355968000000000
        linux_ticks = value.timestamp()
        time_zone = 28800
        linux_ticks = linux_ticks + time_zone
        c_sharp_ticks = linux_ticks * 10000000 + c_sharp_ticks1970
        return int(c_sharp_ticks)



from SapphireQuant.Simulate.DataBlock import DataBlock


class ReplayData:
    """
    播放数据
    """
    def __init__(self, instrument_id):
        self.instrument_id = instrument_id
        self.data_block_map = {}
        self.current_trading_date = None

    def clear(self, trading_date):
        """
        清空指定交易日的缓存
        :param trading_date:
        """
        if trading_date in self.data_block_map.keys():
            tb = self.data_block_map[trading_date]
            tb.data_list.clear()

    def add_data(self, trading_date, data_list):
        """
        添加数据
        :param trading_date:
        :param data_list:
        """
        if trading_date not in self.data_block_map.keys():
            self.data_block_map[trading_date] = DataBlock(data_list)

    def replay_bar_to_time(self, end_time):
        """
        播放到时间
        :param end_time:
        :return:
        """
        if self.current_trading_date not in self.data_block_map.keys():
            return False, []
        tb = self.data_block_map[self.current_trading_date]
        if len(tb.data_list) <= 0:
            return False, []

        i = tb.current_index + 1
        wait_play_bars = []
        while i < len(tb.data_list):
            data = tb.data_list[i]
            if data.end_time <= end_time:
                wait_play_bars.append(data)
            else:
                break
            i += 1

        tb.current_index = i - 1

        if tb.current_index < 0:
            return False, []

        if tb.data_list[i-1].end_time <= end_time:
            return True, wait_play_bars

        return False, []

    def replay_tick_to_time(self, end_time):
        """
        播放到时间
        :param end_time:
        :return:
        """
        if self.current_trading_date not in self.data_block_map.keys():
            return False, []
        tb = self.data_block_map[self.current_trading_date]
        if len(tb.data_list) <= 0:
            return False, []

        i = tb.current_index + 1
        wait_play_ticks = []
        while i < len(tb.data_list):
            data = tb.data_list[i]
            if data.date_time <= end_time:
                wait_play_ticks.append(data)
            else:
                break
            i += 1

        tb.current_index = i - 1

        if tb.current_index < 0:
            return False, []

        if tb.data_list[i-1].date_time <= end_time:
            return True, wait_play_ticks

        return False, []




import datetime

from SapphireQuant.Core.Enum.EnumBarData import EnumBarData


class Bar:
    """
    Bar
    """
    def __init__(self):
        self._instrumentId = ''
        self._tradingDate = datetime.datetime(1990, 1, 1)
        self._beginTime = datetime.datetime(1990, 1, 1)
        self._endTime = datetime.datetime(1990, 1, 1)
        self._high = -100000000.0
        self._open = 0.0
        self._low = 100000000.0
        self._close = 0.0
        self._preClose = 0.0
        self._volume = 0.0
        self._turnover = 0.0
        self._isComplete = False
        self._openInterest = 0.0

    @property
    def instrument_id(self):
        """
        合约编号
        :return:
        """
        return self._instrumentId

    @instrument_id.setter
    def instrument_id(self, value):
        self._instrumentId = value

    @property
    def trading_date(self):
        """
        交易日
        :return:
        """
        return self._tradingDate

    @trading_date.setter
    def trading_date(self, value):
        self._tradingDate = value

    @property
    def begin_time(self):
        """
        开始时间
        :return:
        """
        return self._beginTime

    @begin_time.setter
    def begin_time(self, value):
        self._beginTime = value

    @property
    def end_time(self):
        """
        结束时间
        :return:
        """
        return self._endTime

    @end_time.setter
    def end_time(self, value):
        self._endTime = value

    @property
    def duration(self):
        """
        间隔
        :return:
        """
        return self.end_time - self.begin_time

    @property
    def high(self):
        """
        最高价
        :return:
        """
        if (self._high < 1E-07) & (abs(self.volume) < 1E-10):
            return self._preClose
        return self._high

    @high.setter
    def high(self, value):
        self._high = value

    @property
    def open(self):
        """
        开盘价
        :return:
        """
        if (self._open < 1E-07) & (abs(self.volume) < 1E-10):
            return self._preClose
        return self._open

    @open.setter
    def open(self, value):
        self._open = value

    @property
    def low(self):
        """
        最低价
        :return:
        """
        if (self._low < 1E-07) & (abs(self.volume) < 1E-10):
            return self._preClose
        return self._low

    @low.setter
    def low(self, value):
        self._low = value

    @property
    def close(self):
        """
        收盘价
        :return:
        """
        if (self._close < 1E-07) & (abs(self.volume) < 1E-10):
            return self._preClose
        return self._close

    @close.setter
    def close(self, value):
        self._close = value

    @property
    def pre_close(self):
        """
        前收盘
        :return:
        """
        return self._preClose

    @pre_close.setter
    def pre_close(self, value):
        self._preClose = value

    @property
    def volume(self):
        """
        成交量
        :return:
        """
        return self._volume

    @volume.setter
    def volume(self, value):
        self._volume = value

    @property
    def turnover(self):
        """
        成交额
        :return:
        """
        return self._turnover

    @turnover.setter
    def turnover(self, value):
        self._turnover = value

    @property
    def open_interest(self):
        """
        持仓
        :return:
        """
        return self._openInterest

    @open_interest.setter
    def open_interest(self, value):
        self._openInterest = value

    def __getitem__(self, item):
        if item == EnumBarData.close:
            return self.close
        elif item == EnumBarData.open:
            return self.open
        elif item == EnumBarData.high:
            return self.high
        elif item == EnumBarData.low:
            return self.low
        elif item == EnumBarData.volume:
            return self.volume
        elif item == EnumBarData.turnover:
            return self.turnover
        else:
            return 0.0

    @property
    def bar_change(self):
        """
        涨跌幅:close/open - 1.0
        :return:
        """
        num = 0.0
        if abs(self.open) > 1E-10:
            num = round(100.0 * (self.close / self.open - 1.0), 3)
        return num

    @property
    def change(self):
        """
        涨跌幅:close/pre_close - 1.0
        :return:
        """
        num = 0.0
        if abs(self.pre_close) > 1E-10:
            num = round(100.0 * (self.close / self.pre_close - 1.0), 3)
        return num

    def add_tick(self, tick):
        """
        添加tick
        :param tick:
        """
        if self._high < tick.last_price:
            self._high = tick.last_price
        if self._low > tick.last_price:
            self._low = tick.last_price
        self._close = tick.last_price
        self.end_time = tick.date_time
        self.open_interest = tick.open_interest

    def add_bar(self, bar):
        """
        添加bar
        :param bar:
        """
        if self._high < bar.high:
            self._high = bar.high
        if self._low > bar.low:
            self._low = bar.low
        self._close = bar.close
        self.end_time = bar.end_time
        self.open_interest = bar.open_interest
        self.volume += bar.volume
        self.turnover += bar.turnover
        self.trading_date = bar.trading_date

    def open_bar(self, begin_time, tick, bar):
        """
        开始一个bar
        :param begin_time:
        :param tick:
        :param bar:
        """
        self.begin_time = begin_time
        self._high = self._low = self._open = self._close = tick.last_price
        if bar is None:
            if (tick.high_price > 0.0) & (tick.high_price < 10000000000.0):
                self.high = tick.high_price
            if (tick.low_price > 0.0) & (tick.low_price < 10000000000.0):
                self.low = tick.low_price
            if (tick.open_price > 0.0) & (tick.open_price < 10000000000.0):
                self.open = tick.open_price

        self.end_time = tick.date_time
        if bar is not None:
            self.pre_close = bar.close
        else:
            if tick.pre_close_price > 0.0:
                self.pre_close = tick.pre_close_price
            else:
                self.pre_close = tick.open_price

        self.open_interest = tick.open_interest

    def open_bar_with_new_bar(self, new_bar):
        """
        用新的bar
        :param new_bar:
        """
        self.instrument_id = new_bar.instrument_id
        self.begin_time = new_bar.begin_time
        self.high = new_bar.high
        self.low = new_bar.low
        self.open = new_bar.open
        self.close = new_bar.close
        self.end_time = new_bar.end_time
        self.volume = new_bar.volume
        self.turnover = new_bar.turnover
        self.pre_close = new_bar.pre_close
        self.open_interest = new_bar.open_interest

    def close_bar(self, end_time):
        """
        结束一个bar
        :param end_time:
        """
        self.end_time = end_time

    def clone(self):
        """
        克隆
        :return:
        """
        bar = Bar()
        bar._instrumentId = self._instrumentId
        bar._beginTime = self._beginTime
        bar._endTime = self._endTime
        bar._high = self._high
        bar._open = self._open
        bar._low = self._low
        bar._close = self._close
        bar._preClose = self._preClose
        bar._volume = self._volume
        bar._turnover = self._turnover
        bar._isComplete = self._isComplete
        bar._openInterest = self._openInterest
        bar._tradingDate = self._tradingDate
        return bar

    def to_string(self):
        """
        转成string
        :return:
        """
        msg = "[{0}]".format(self._instrumentId) + "[" + str(self.begin_time) + "-" + str(self.end_time) + "]"
        msg += "高" + str(self.high)
        msg += "开" + str(self.open)
        msg += "低" + str(self.low)
        msg += "收" + str(self.close)
        msg += "量=" + str(self.volume)
        msg += "额=" + str(self.turnover)
        msg += "持仓" + str(self.open_interest)
        msg += "交易日=" + self.trading_date.strftime("%Y/%m/%d")

        return msg



from SapphireQuant.Core.Entity import Entity
from SapphireQuant.Core.Enum import EnumPositionDirection, EnumHedgeFlag


class FuturePosition(Entity):
    """
    期货持仓
    """

    def __init__(self):
        super().__init__()
        self.instrument_id = ''
        self.instrument_name = ''
        self.direction = EnumPositionDirection.净持仓
        self.hedge_flag = EnumHedgeFlag.套保
        self.exchange_id = ''
        self.last_price = 0
        self.volume = 0
        self.td_volume = 0
        self.yd_volume = 0
        self.td_open_trade = 0
        self.td_close_trade = 0
        self.open_cost = 0
        self.open_volume = 0
        self.close_volume = 0
        self.accumulated_close_profit = 0
        self.transaction_cost = 0
        self.close_profit = 0
        self.position_profit = 0
        self.market_value = 0
        self.num_holding_day = 0
        self.margin = 0

    def clone(self):
        position = FuturePosition()
        position.product_code = self.product_code
        position.strategy_id = self.strategy_id
        position.instrument_id = self.instrument_id
        position.instrument_name = self.instrument_name
        position.direction = self.direction
        position.hedge_flag = self.hedge_flag
        position.exchange_id = self.exchange_id
        position.last_price = self.last_price
        position.volume = self.volume
        position.td_volume = self.td_volume
        position.yd_volume = self.yd_volume
        position.td_open_trade = self.td_open_trade
        position.td_close_trade = self.td_close_trade
        position.open_cost = self.open_cost
        position.open_volume = self.open_volume
        position.close_volume = self.close_volume
        position.accumulated_close_profit = self.accumulated_close_profit
        position.transaction_cost = self.transaction_cost
        position.close_profit = self.close_profit
        position.position_profit = self.position_profit
        position.market_value = self.market_value
        position.margin = self.margin

        return position

    def to_string(self):
        """

        :return:
        """
        return '{0}:{1},总仓:{2},昨仓:{3},今仓:{4},持仓均价:{5},平仓盈亏:{6},持仓盈亏:{7},占用保证金:{8},持仓市值:{9}'.format(
            self.instrument_id, self.direction.chinese_name(), self.volume, self.yd_volume, self.td_volume, self.open_cost, self.close_profit, self.position_profit, self.margin,
            self.market_value)



from enum import Enum, unique


@unique
class EnumPositionType(Enum):
    今仓 = 0
    昨仓 = 1

    def chinese_name(self):
        """

        :return:
        """
        if self == EnumPositionType.今仓:
            return '今仓'
        else:
            return '昨仓'



import datetime
import os
import os.path

from SapphireQuant import BarSeries
from SapphireQuant.Core.Bar import Bar
from SapphireQuant.Core.CSharpUtils import CSharpUtils
from SapphireQuant.Core.Enum.EnumBarType import EnumBarType
from SapphireQuant.DataProcessing.BinaryWriter import BinaryWriter
from SapphireQuant.DataProcessing.BinaryReader import BinaryReader
from SapphireQuant.DataProcessing.FileHeader import FileHeader


class BinaryMinBarStream(object):
    """
    分钟读写流
    """
    BarLength = 88
    SimpleDayBarCountThreshold = 18

    def __init__(self, market, instrument_id, file_path):
        self._file_extension = ".min"
        self._instrument_id = instrument_id.split('.')[0]
        self._file_path = file_path
        self._market = market
        self._file_version = 'MQ 1.0'

        if file_path.strip() == '':
            raise Exception("Invalid Config File Path!")

        extension = os.path.splitext(file_path)[1].lower()
        if extension != self._file_extension:
            raise Exception("Invalid File Extension!")

    def _read_bar(self, reader):
        """
        读取Bar
        :param reader:
        :return:
        """
        bar = Bar()
        bar.begin_time = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())
        bar.end_time = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())
        bar.trading_date = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())
        bar.open = reader.read_double()
        bar.close = reader.read_double()
        bar.high = reader.read_double()
        bar.low = reader.read_double()
        bar.pre_close = reader.read_double()
        bar.volume = reader.read_double()
        bar.turnover = reader.read_double()
        bar.open_interest = reader.read_double()
        bar.instrument_id = self._instrument_id
        # bar.IsCompleted = True

        return bar

    def read_all(self, bar_series):
        """

        :param bar_series:
        :return:
        """
        if bar_series is None:
            raise Exception("BarSeries Is Empty.")

        with open(self._file_path, 'rb+') as f:
            reader = BinaryReader(f)
            file_header = FileHeader()
            file_header.read(reader)
            begin_date = file_header.begin_trading_day
            end_date = file_header.end_trading_day

            return self.read_trading_days(bar_series, begin_date, end_date)

    def read_trading_day(self, bar_series, trading_day):
        """
        按照交易日读取分钟线数据
        :param bar_series:分钟线数据
        :param trading_day:交易日
        :return:
        """
        if bar_series is None:
            raise Exception("BarSeries Is Empty.")

        if not isinstance(trading_day, datetime.date):
            raise Exception("Trading Day Param Error")

        read_count = 0
        trading_day = datetime.datetime(trading_day.year, trading_day.month, trading_day.day)
        with open(self._file_path, 'rb+') as f:
            reader = BinaryReader(f)
            file_header = FileHeader()
            file_header.read(reader)
            if (trading_day < file_header.begin_trading_day) or (trading_day > file_header.end_trading_day):
                return False

            data = file_header.trading_day_indices.try_get(trading_day)
            found = data[0]
            start_offset = data[1]
            count = data[2]
            if found:
                reader.stream.seek(start_offset, 0)
                while count > 0:
                    bar = self._read_bar(reader)
                    bar_series.append(bar)
                    count -= 1
                    read_count += 1

            return read_count > 0

    def read_trading_days(self, bar_series, begin_trading_day, end_trading_day):
        """
        按照交易日区间读取分钟线数据
        :param bar_series:分钟线数据
        :param begin_trading_day:开始交易日
        :param end_trading_day:结束交易日
        :return:
        """
        if bar_series is None:
            raise Exception("BarSeries Is Empty.")

        if not isinstance(begin_trading_day, datetime.date):
            raise Exception("BeginTradingDay Param Error.")

        if not isinstance(end_trading_day, datetime.date):
            raise Exception("EndTradingDay Param Error.")

        begin_trading_day = datetime.datetime(begin_trading_day.year, begin_trading_day.month, begin_trading_day.day)
        end_trading_day = datetime.datetime(end_trading_day.year, end_trading_day.month, end_trading_day.day)

        if begin_trading_day > end_trading_day:
            raise Exception("EndTradingDay Must Be Larger Than BeginTradingDay")

        if begin_trading_day == end_trading_day:
            return self.read_trading_day(bar_series, begin_trading_day)

        read_count = 0
        with open(self._file_path, 'rb+') as f:
            reader = BinaryReader(f)
            file_header = FileHeader()
            file_header.read(reader)
            if (end_trading_day < file_header.begin_trading_day) or (
                    begin_trading_day > file_header.end_trading_day):
                return False

            if begin_trading_day < file_header.begin_trading_day:
                begin_trading_day = file_header.begin_trading_day

            if end_trading_day > file_header.end_trading_day:
                end_trading_day = file_header.end_trading_day

            data = file_header.trading_day_indices.get_trading_day_offsets(begin_trading_day, end_trading_day)
            found = data[0]
            start_offset = data[1]
            end_offset = data[2]
            start_count = data[3]
            end_count = data[4]
            if found:
                end_offset += self.BarLength * end_count
                reader.stream.seek(start_offset, 0)
                while reader.stream.tell() < end_offset:
                    bar = self._read_bar(reader)
                    bar_series.append(bar)
                    read_count += 1

            return read_count > 0

    @staticmethod
    def _update_file_header(file_header, bar_series, start, length):
        start_bar = bar_series[start]
        end_bar = bar_series[start + length - 1]
        if file_header.begin_time == datetime.datetime(2999, 1, 1) or start_bar.begin_time < file_header.begin_time:
            file_header.begin_time = start_bar.begin_time
            file_header.mark_as_dirty(True)

        if file_header.end_time == datetime.datetime(2999, 1, 1) or start_bar.end_time > file_header.end_time:
            file_header.end_time = end_bar.end_time
            file_header.mark_as_dirty(True)
        file_header.bar_count += length

    def _write_interval(self, file_header, bar_series: BarSeries, start: int, length: int, offset: int):
        if bar_series is None:
            raise Exception("Argument Null Exception:bar_series")
        if length < 0:
            raise Exception("Argument Null Exception:length")
        if start < 0 or (start + length > len(bar_series)):
            raise Exception("Argument Null Exception:start")

        with open(self._file_path, 'rb+') as f:
            writer = BinaryWriter(f)
            BinaryMinBarStream._update_file_header(file_header, bar_series, start, length)
            BinaryMinBarStream._write_bars(writer, file_header, bar_series, start, length, offset)
            file_header.write(writer)

    @staticmethod
    def _write_bars(writer, file_header, bar_series, start, length, offset):
        writer.stream.seek(offset, 0)
        for i in range(start, length, 1):
            bar = bar_series[i]
            BinaryMinBarStream._write_bar(writer, file_header, bar)

    @staticmethod
    def _write_bar(writer, file_header, bar):
        trading_date = datetime.datetime(bar.trading_date.year, bar.trading_date.month, bar.trading_date.day)
        if file_header.begin_trading_day == datetime.datetime(2999, 1, 1) or trading_date < file_header.begin_trading_day:
            file_header.begin_trading_day = trading_date
            file_header.mark_as_dirty(True)
        if file_header.end_trading_day == datetime.datetime(2999, 1, 1) or trading_date > file_header.end_trading_day:
            file_header.end_trading_day = trading_date
            file_header.mark_as_dirty(True)
        offset = writer.stream.tell()
        file_header.natural_day_indices.add(datetime.datetime(bar.begin_time.year, bar.begin_time.month, bar.begin_time.day), offset)
        file_header.trading_day_indices.add(bar.trading_date, offset)
        writer.write_int64(CSharpUtils.get_c_sharp_ticks(bar.begin_time))
        writer.write_int64(CSharpUtils.get_c_sharp_ticks(bar.end_time))
        writer.write_int64(CSharpUtils.get_c_sharp_ticks(bar.trading_date))
        writer.write_double(bar.open)
        writer.write_double(bar.close)
        writer.write_double(bar.high)
        writer.write_double(bar.low)
        writer.write_double(bar.pre_close)
        writer.write_double(bar.volume)
        writer.write_double(bar.turnover)
        writer.write_double(bar.open_interest)

    def write(self, bar_series):
        self.write_bar_series(bar_series, 0, len(bar_series))

    def write_trading_date(self, bar_series, trading_date):
        create_new = not os.path.exists(self._file_path)
        if create_new:
            self.write_bar_series(bar_series, 0, len(bar_series))
        else:
            with open(self._file_path, 'rb+') as f:
                reader = BinaryReader(f)
                file_header = FileHeader()
                file_header.read(reader)
                data = file_header.trading_day_indices.try_get(trading_date)
                found = data[0]
                if found:
                    start_offset = data[1]
                    count = data[2]
                    file_header.trading_day_indices.reset_count(trading_date)
                    file_header.natural_day_indices.reset_count(trading_date)
                    file_header.bar_count = file_header.trading_day_indices.get_count()
                    file_header.mark_as_dirty(True)
                    self._write_interval(file_header, bar_series, 0, len(bar_series), start_offset)
                else:
                    reader.stream.seek(0, 2)
                    start_offset = reader.stream.tell()
                    self._write_interval(file_header, bar_series, 0, len(bar_series), start_offset)

    def write_bar_series(self, bar_series: BarSeries, start: int, length: int):
        create_new = not os.path.exists(self._file_path)

        file_header = None
        if create_new:
            record_deep_dir = os.path.dirname(self._file_path)
            if not os.path.exists(record_deep_dir):
                os.makedirs(record_deep_dir)

            file_stream = open(self._file_path, 'w')
            file_stream.close()

            file_header = FileHeader()
            file_header.file_version = self._file_version
            file_header.period = 1
            file_header.market = self._market
            file_header.bar_type = EnumBarType.minute
        else:
            if os.path.getsize(self._file_path) == 0:
                raise Exception("Invalid File Length!")

        self._write_interval(file_header, bar_series, start, length, FileHeader.FileHeaderTotalSize)



from SapphireQuant.Config.Ctp.CtpCounter import CtpCounter
from SapphireQuant.Config.Ctp.EnumNetworkType import EnumNetworkType


class CtpCounterGroup:
    """
    Ctp柜台组
    """

    def __init__(self, name):
        self.name = name
        self.telecom_counter = None
        self.unicom_counter = None

    def add(self, server):
        """
        添加服务器
        :param server:
        """
        if server.network_type == EnumNetworkType.telecom:
            if self.telecom_counter is None:
                self.telecom_counter = CtpCounter()
                self.telecom_counter.add(server)
        if server.network_type == EnumNetworkType.unicom:
            if self.unicom_counter is None:
                self.unicom_counter = CtpCounter()
                self.unicom_counter.add(server)

    def get_counter(self, network_type):
        """
        获取柜台
        :param network_type:
        :return:
        """
        if network_type == EnumNetworkType.telecom:
            return self.telecom_counter
        if network_type == EnumNetworkType.unicom:
            return self.unicom_counter
        return None

    def get_servers(self, network_type, server_type):
        """
        获取服务器
        :param network_type:
        :param server_type:
        :return:
        """
        if network_type == EnumNetworkType.telecom:
            return self.telecom_counter.get_servers(network_type, server_type)
        if network_type == EnumNetworkType.unicom:
            return self.unicom_counter.get_servers(network_type, server_type)
        return None



import datetime
import os
import sys
from SapphireQuant.Core.CSharpUtils import CSharpUtils
from SapphireQuant.Core.Quote import Quote
from SapphireQuant.Core.Tick import Tick
from SapphireQuant.DataProcessing.BinaryReader import BinaryReader
from SapphireQuant.DataProcessing.BinaryWriter import BinaryWriter


class BinaryDcTickStream:
    """
    Tick读写流
    """
    CFileFlag = 1262700884  # ('K' << 24) + ('C' << 16) + ('I' << 8) + 'T'
    CFileHeaderLen = 58

    CTickHeaderLenV1 = 64  # 带LocalTime
    CTickQuoteLen = 5 * 32  # 买一价，卖一价，买一量，卖一量

    def __init__(self, market, instrument_id, exchange_id, file_path):
        self._file_path = file_path
        self._instrument_id = instrument_id
        self._exchange_id = exchange_id
        self._market = market

        # Header
        self._version = 1
        self._trading_day = datetime.datetime.max
        self._pre_close_price = 0.0
        self._pre_settlement_price = 0.0
        self._pre_interest = 0.0
        self._up_limit = 0.0
        self._down_limit = 0.0
        self._open_price = 0.0

    def _read_header(self, reader):
        """
        读取头
        :param reader:
        :return:
        """
        reader.stream.seek(0, 0)
        flag = reader.read_int32()
        if flag != self.CFileFlag:
            return False
        self._version = reader.read_int16()

        year = reader.read_int16()
        month = reader.read_byte()
        day = reader.read_byte()

        self._trading_day = datetime.datetime(year, month, day)

        self._pre_close_price = reader.read_double()
        self._pre_settlement_price = reader.read_double()
        self._pre_interest = reader.read_double()
        self._up_limit = reader.read_double()
        self._down_limit = reader.read_double()
        self._open_price = reader.read_double()
        return True

    def _read_tick(self, reader):
        """
        读取Tick
        :param reader:
        """
        tick = Tick()
        tick.market = self._market
        tick.open_price = self._open_price
        tick.pre_close_price = self._pre_close_price
        tick.instrument_id = self._instrument_id
        tick.exchange_id = self._exchange_id
        tick.pre_open_interest = self._pre_interest
        tick.pre_settlement_price = self._pre_settlement_price
        tick.up_limit = self._up_limit
        tick.drop_limit = self._down_limit

        tick.date_time = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())
        tick.local_time = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())
        tick.trading_date = self._trading_day
        tick.last_price = reader.read_double()
        tick.high_price = reader.read_double()
        tick.low_price = reader.read_double()
        tick.open_interest = reader.read_double()
        tick.volume = reader.read_double()
        tick.turnover = reader.read_double()
        quote = Quote()
        quote.ask_volume1 = reader.read_double()
        quote.bid_volume1 = reader.read_double()
        quote.ask_price1 = reader.read_double()
        quote.bid_price1 = reader.read_double()

        quote.ask_volume2 = reader.read_double()
        quote.bid_volume2 = reader.read_double()
        quote.ask_price2 = reader.read_double()
        quote.bid_price2 = reader.read_double()

        quote.ask_volume3 = reader.read_double()
        quote.bid_volume3 = reader.read_double()
        quote.ask_price3 = reader.read_double()
        quote.bid_price3 = reader.read_double()

        quote.ask_volume4 = reader.read_double()
        quote.bid_volume4 = reader.read_double()
        quote.ask_price4 = reader.read_double()
        quote.bid_price4 = reader.read_double()

        quote.ask_volume5 = reader.read_double()
        quote.bid_volume5 = reader.read_double()
        quote.ask_price5 = reader.read_double()
        quote.bid_price5 = reader.read_double()
        tick.quote = quote
        return tick

    def _get_tick_count(self):
        tick_count = int((os.path.getsize(self._file_path) - self.CFileHeaderLen) / (self.CTickHeaderLenV1 + self.CTickQuoteLen))
        return tick_count

    def _read_ticks_by_count(self, tick_series, reader, offset, count):
        """
        读取V1版本的Level1行情
        :param tick_series:
        :param reader:
        :param offset:
        :param count:
        :return:
        """
        pos = self.CFileHeaderLen + offset * (self.CTickHeaderLenV1 + self.CTickQuoteLen)
        reader.stream.seek(pos)
        tick_count = self._get_tick_count()
        read_tick_count = tick_count - offset
        if read_tick_count < count:
            read_tick_count = read_tick_count
        else:
            read_tick_count = count

        for i in range(0, read_tick_count):
            tick = self._read_tick(reader)
            tick_series.append(tick)

    def _read_ticks_by_time(self, tick_series, reader, begin_time, end_time):
        """

        :param tick_series:
        :param reader:
        :param begin_time:
        :param end_time:
        :return:
        """
        reader.stream.seek(self.CFileHeaderLen, 0)
        tick_count = self._get_tick_count()
        blk_len = self.CTickHeaderLenV1 + self.CTickQuoteLen - 8
        for i in range(tick_count):
            date_time = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(reader.read_int64())

            if date_time > end_time:
                return

            if date_time < begin_time:
                reader.read_bytes(blk_len)
                continue

            tick = self._read_tick(reader)
            tick_series.append(tick)

    def read_last_tick(self):
        """
        根据个数读取
        :return:
        """
        if not os.path.exists(self._file_path):
            _firstRead = True
            print(f"Read Future Tick Data Failed:{self._file_path},File Do Not Exist....")
            return None

        with open(self._file_path, 'rb') as stream:
            reader = BinaryReader(stream)
            self._read_header(reader)
            tick_count = self._get_tick_count()
            tick_series = []
            offset = tick_count - 1
            count = 1
            self._read_ticks_by_count(tick_series, reader, offset, count)

            if len(tick_series) > 0:
                return tick_series[0]
        return None

    def read_by_count(self, tick_series, offset, count):
        """
        根据个数读取
        :param tick_series:
        :param offset:跳多少个Tick去取数据
        :param count:
        :return:
        """
        if not os.path.exists(self._file_path):
            print("Read Tick Data Error,Do Not Exist File:{0}".format(self._file_path))
            return False

        with open(self._file_path, 'rb') as stream:
            reader = BinaryReader(stream)
            self._read_header(reader)
            self._read_ticks_by_count(tick_series, reader, offset, count)
            return True

    def read_by_time(self, tick_series, begin_time, end_time):
        """
        根据时间读取Tick
        :param tick_series:
        :param begin_time:
        :param end_time:
        :return:
        """
        if (begin_time is None) and (end_time is None):
            return self.read_by_count(tick_series, 0, sys.maxsize)

        if end_time is None:
            end_time = datetime.datetime.max

        try:
            if not os.path.exists(self._file_path):
                print(f"Read Tick Data Error,Do Not Exist File:{self._file_path}")
                return False

            with open(self._file_path, 'rb') as stream:
                reader = BinaryReader(stream)
                self._read_header(reader)
                self._read_ticks_by_time(tick_series, reader, begin_time, end_time)
            return True
        except Exception as e:
            print("Read Tick Data By Time Error:{0}".format(str(e)))

        return False

    def _write_header(self, writer):
        """
        写头
        :param writer:
        """
        writer.stream.seek(0, 0)
        writer.write_int32(self.CFileFlag)
        writer.write_int16(self._version)
        writer.write_int16(self._trading_day.year)
        writer.write_byte(self._trading_day.month)
        writer.write_byte(self._trading_day.day)
        writer.write_double(self._pre_close_price)
        writer.write_double(self._pre_settlement_price)
        writer.write_double(self._pre_interest)
        writer.write_double(self._up_limit)
        writer.write_double(self._down_limit)
        writer.write_double(self._open_price)

    def _write_ticks(self, tick_list, writer, offset, count):
        """
        写入
        """
        end = offset + count
        for i in range(offset, end, 1):
            if i >= end:
                break
            tick = tick_list[i]
            writer.write_int64(CSharpUtils.get_c_sharp_ticks(tick.date_time))
            writer.write_int64(CSharpUtils.get_c_sharp_ticks(tick.local_time))

            writer.write_double(tick.last_price)
            writer.write_double(tick.high_price)
            writer.write_double(tick.low_price)
            writer.write_double(tick.open_interest)

            writer.write_double(tick.volume)
            writer.write_double(tick.turnover)

            writer.write_double(tick.quote.ask_volume1)
            writer.write_double(tick.quote.bid_volume1)
            writer.write_double(tick.quote.ask_price1)
            writer.write_double(tick.quote.bid_price1)

            writer.write_double(tick.quote.ask_volume2)
            writer.write_double(tick.quote.bid_volume2)
            writer.write_double(tick.quote.ask_price2)
            writer.write_double(tick.quote.bid_price2)

            writer.write_double(tick.quote.ask_volume3)
            writer.write_double(tick.quote.bid_volume3)
            writer.write_double(tick.quote.ask_price3)
            writer.write_double(tick.quote.bid_price3)

            writer.write_double(tick.quote.ask_volume4)
            writer.write_double(tick.quote.bid_volume4)
            writer.write_double(tick.quote.ask_price4)
            writer.write_double(tick.quote.bid_price4)

            writer.write_double(tick.quote.ask_volume5)
            writer.write_double(tick.quote.bid_volume5)
            writer.write_double(tick.quote.ask_price5)
            writer.write_double(tick.quote.bid_price5)

    def write(self, tick_list, start, length):
        """

        :param tick_list:
        :param start:
        :param length:
        :return:
        """
        if len(tick_list) == 0:
            return True

        try:
            self._trading_day = tick_list[0].trading_date
            self._pre_close_price = tick_list[0].pre_close_price
            self._open_price = tick_list[0].open_price
            self._pre_interest = tick_list[0].pre_open_interest
            self._pre_settlement_price = tick_list[0].pre_settlement_price
            self._up_limit = tick_list[0].up_limit
            self._down_limit = tick_list[0].drop_limit

            count = len(tick_list) - start
            if length < count:
                count = length

            with open(self._file_path, 'wb') as stream:
                writer = BinaryWriter(stream)
                self._write_header(writer)
                self._write_ticks(tick_list, writer, start, count)
                stream.flush()

            return True
        except Exception as e:
            print('Write Ticks Error:{0}'.format(str(e)))

    def append(self, tick_list, start, length):
        """
        追加Tick
        :param tick_list:
        :param start:
        :param length:
        :return:
        """
        if not os.path.exists(self._file_path):
            print('Append Tick, File Not Exist, Cover:{0}'.format(self._file_path))
            record_deep_dir = os.path.dirname(self._file_path)
            if not os.path.exists(record_deep_dir):
                os.makedirs(record_deep_dir)
            return self.write(tick_list, start, length)
        if os.path.getsize(self._file_path) < self.CFileHeaderLen:
            print(f'Append Tick, File Size Error, Cover:{self._file_path}')
            return self.write(tick_list, start, length)
        if (os.path.getsize(self._file_path) - self.CFileHeaderLen) % (self.CTickHeaderLenV1 + self.CTickQuoteLen) != 0:
            print(f'Append Tick, File Layout Error, Cover:{self._file_path}')
            return self.write(tick_list, start, length)
        try:
            # with open(self._file_path, 'rb') as stream:
            #     reader = BinaryReader(stream)
            #     self._read_header(reader)

            with open(self._file_path, 'rb+') as stream:
                count = len(tick_list) - start
                if length < count:
                    count = length

                writer = BinaryWriter(stream)
                stream.seek(0, 2)
                self._write_ticks(tick_list, writer, start, count)

                stream.flush()
            return True

        except Exception as e:
            print("Append Tick Error:{0},{1}".format(self._file_path, str(e)))
        return False



from .EnumBarData import EnumBarData
from .EnumBarType import EnumBarType
from .EnumBuySell import EnumBuySell
from .EnumDriveMode import EnumDriveMode
from .EnumExchange import EnumExchange
from .EnumFutureType import EnumFutureType
from .EnumHedgeFlag import EnumHedgeFlag
from .EnumLifePhase import EnumLifePhase
from .EnumMarket import EnumMarket
from .EnumOpenClose import EnumOpenClose
from .EnumOrderPriceType import EnumOrderPriceType
from .EnumOrderRejectReason import EnumOrderRejectReason
from .EnumOrderStatus import EnumOrderStatus
from .EnumOrderSubmitStatus import EnumOrderSubmitStatus
from .EnumOrderType import EnumOrderType
from .EnumPositionDirection import EnumPositionDirection
from .EnumPositionType import EnumPositionType
from .EnumQuoteEventType import EnumQuoteEventType
from .EnumProductClass import EnumProductClass
from .EnumQuoteEventType import EnumQuoteEventType
from .EnumRestoration import EnumRestoration
from .EnumSimulateMatchMode import EnumSimulateMatchMode
from .EnumStrategyWorkMode import EnumStrategyWorkMode
from .EnumTradeEventType import EnumTradeEventType


__all__ = ['EnumBarData',
           'EnumBarType',
           'EnumBuySell',
           'EnumDriveMode',
           'EnumExchange',
           'EnumFutureType',
           'EnumHedgeFlag',
           'EnumLifePhase',
           'EnumMarket',
           'EnumOpenClose',
           'EnumOrderPriceType',
           'EnumOrderRejectReason',
           'EnumOrderStatus',
           'EnumOrderSubmitStatus',
           'EnumOrderType',
           'EnumPositionDirection',
           'EnumPositionType',
           'EnumProductClass',
           'EnumQuoteEventType',
           'EnumRestoration',
           'EnumSimulateMatchMode',
           'EnumStrategyWorkMode',
           'EnumTradeEventType',
           ]



from enum import Enum, unique


@unique
class EnumHedgeFlag(Enum):
    投机 = 0
    套利 = 1
    套保 = 2

    def chinese_name(self):
        """

        :return:
        """
        if self == EnumHedgeFlag.投机:
            return '投机'
        elif self == EnumHedgeFlag.套利:
            return '套利'
        else:
            return '套保'



import logging
import os


def create_log(log_dir=None, log_name=None):
    """

    :param log_dir:
    :param log_name:
    :return:
    """
    if log_dir is None:
        log_dir = "log_dir"

    path = os.path.join(os.path.abspath("."), log_dir)

    if not os.path.exists(path):
        os.makedirs(path)
        print('创建日志目录:{0}'.format(path))

    logger = logging.getLogger(log_name)
    # 配置日志信息
    if log_name is None:
        log_name = "demo.log"
    log_name = os.path.join(path, log_name)

    # 创建一个日志 logger 实例
    logger.setLevel(logging.DEBUG)  # 设置日志级别为 DEBUG， 覆盖掉默认的日志级别 Warning

    # 创建一个 handler，用于写入日志文件， handler 可以把日志写到不同的地方
    print('创建日志文件:{0}'.format(log_name))
    fh = logging.FileHandler(log_name, "w+", encoding="utf-8")  # 将日志写在文件中
    fh.setLevel(logging.DEBUG)  # 设置日志的级别为  INFO

    # 再创建一个 handler, 用于输出控制台
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)  # 设置日志的级别为 INFO

    # 定义handler的格式输出
    log_format = logging.Formatter("%(asctime)s - %(filename)s - %(levelname)s - %(name)s - %(message)s")  # 时间，文件名称，日志级别， 日志器名称， 字段信息

    fh.setFormatter(log_format)  # handler 加载设置的格式输出
    ch.setFormatter(log_format)  # handler 加载设置的格式输出

    # 为 logger 添加handler
    logger.addHandler(fh)
    logger.addHandler(ch)
    return logger




from abc import ABC

from SapphireQuant.Strategy.BaseStrategy import BaseStrategy


class MixStrategy(BaseStrategy, ABC):
    """
    混合策略基类
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.strategy_series = []

    def add_strategy(self, strategy):
        """

        :param strategy:
        """
        self.strategy_series.append(strategy)
        key = 'ID_{0}_{1}'.format(strategy.product_code, strategy.strategy_id)
        if key in self.kwargs.keys():
            print('开始初始化策略:{0}, {1}'.format(strategy.product_code, strategy.strategy_id))
            strategy.__init__(self.kwargs[key])

    def on_start(self):
        """
        开始
        """
        for strategy in self.strategy_series:
            strategy.on_start()

    def on_bar(self, bar):
        """

        :param bar:
        """
        for strategy in self.strategy_series:
            strategy.on_bar(bar)

    def on_tick(self, tick):
        """

        :param tick:
        """
        for strategy in self.strategy_series:
            strategy.on_tick(tick)

    def on_exit(self):
        """
        策略结束
        """
        for strategy in self.strategy_series:
            strategy.on_exit()



from .Account import *
from .Ctp import *
from .Instrument import *
from .TradingDay import *
from .TradingFrame import *

__all__ = []
__all__.extend(Account.__all__)
__all__.extend(Ctp.__all__)
__all__.extend(Instrument.__all__)
__all__.extend(TradingDay.__all__)
__all__.extend(TradingFrame.__all__)



import time

from SapphireQuant.Core.Quote import Quote


class DcTick:
    """
    Tick数据
    """

    def __init__(self):
        self._instrument_id = ""
        self._exchange_id = ""
        self._quote = Quote()
        self._last_price = 0.0
        self._iopv = 0.0
        self._pre_close_price = 0.0
        self._open_price = 0.0
        self._high_price = 0.0
        self._low_price = 0.0
        self._volume = 0.0
        self._turnover = 0.0
        self._up_limit_price = 0.0
        self._drop_limit_price = 0.0
        self._trading_date = time.time()
        self._natural_time = time.time()
        self._local_time = time.time()
        self._open_interest = 0.0
        self._pre_open_interest = 0.0
        self._pre_settlement_price = 0.0
        self._market = 1
        self._status = 1
        self.bid_price_level = 1
        self.ask_price_level = 1

        self.num_trades = 0
        self.total_bid_qty = 0
        self.total_ask_qty = 0
        self.weighted_avg_bid_price = 0.0
        self.weighted_avg_ask_price = 0.0

    @property
    def instrument_id(self):
        """

        :return:
        """
        return self._instrument_id

    @instrument_id.setter
    def instrument_id(self, value):
        self._instrument_id = value

    @property
    def exchange_id(self):
        """

        :return:
        """
        return self._exchange_id

    @exchange_id.setter
    def exchange_id(self, value):
        self._exchange_id = value

    @property
    def market(self):
        """

        :return:
        """
        return self._market

    @market.setter
    def market(self, value):
        self._market = value

    @property
    def local_time(self):
        """

        :return:
        """
        return self._local_time

    @local_time.setter
    def local_time(self, value):
        self._local_time = value

    @property
    def date_time(self):
        """

        :return:
        """
        return self._natural_time

    @date_time.setter
    def date_time(self, value):
        self._natural_time = value

    @property
    def time_now(self):
        """

        :return:
        """
        return time.mktime(str(self._natural_time.timetuple()))

    @property
    def trading_date(self):
        """

        :return:
        """
        return self._trading_date

    @trading_date.setter
    def trading_date(self, value):
        self._trading_date = value

    @property
    def quote(self):
        """

        :return:
        """
        return self._quote

    @quote.setter
    def quote(self, value):
        self._quote = value

    @property
    def ask_price1(self):
        """

        :return:
        """
        return self._quote.ask_price1

    @ask_price1.setter
    def ask_price1(self, value):
        self._quote.ask_price1 = value

    @property
    def ask_volume1(self):
        """

        :return:
        """
        return self._quote.ask_volume1

    @ask_volume1.setter
    def ask_volume1(self, value):
        self._quote.ask_volume1 = value

    @property
    def bid_price1(self):
        """

        :return:
        """
        return self._quote.bid_price1

    @bid_price1.setter
    def bid_price1(self, value):
        self._quote.bid_price1 = value

    @property
    def bid_volume1(self):
        """

        :return:
        """
        return self._quote.bid_volume1

    @bid_volume1.setter
    def bid_volume1(self, value):
        self._quote.bid_volume1 = value

    @property
    def pre_close_price(self):
        """

        :return:
        """
        return self._pre_close_price

    @pre_close_price.setter
    def pre_close_price(self, value):
        self._pre_close_price = value

    @property
    def open_price(self):
        """

        :return:
        """
        return self._open_price

    @open_price.setter
    def open_price(self, value):
        self._open_price = value

    @property
    def high_price(self):
        """

        :return:
        """
        return self._high_price

    @high_price.setter
    def high_price(self, value):
        self._high_price = value

    @property
    def low_price(self):
        """

        :return:
        """
        return self._low_price

    @low_price.setter
    def low_price(self, value):
        self._low_price = value

    @property
    def last_price(self):
        """

        :return:
        """
        return self._last_price

    @last_price.setter
    def last_price(self, value):
        self._last_price = value

    @property
    def iopv(self):
        """

        :return:
        """
        return self._iopv

    @iopv.setter
    def iopv(self, value):
        self._iopv = value

    @property
    def volume(self):
        """

        :return:
        """
        return self._volume

    @volume.setter
    def volume(self, value):
        self._volume = value

    @property
    def turnover(self):
        """

        :return:
        """
        return self._turnover

    @turnover.setter
    def turnover(self, value):
        self._turnover = value

    @property
    def up_limit(self):
        """

        :return:
        """
        return self._up_limit_price

    @up_limit.setter
    def up_limit(self, value):
        self._up_limit_price = value

    @property
    def drop_limit(self):
        """

        :return:
        """
        return self._drop_limit_price

    @drop_limit.setter
    def drop_limit(self, value):
        self._drop_limit_price = value

    @property
    def open_interest(self):
        """

        :return:
        """
        return self._open_interest

    @open_interest.setter
    def open_interest(self, value):
        self._open_interest = value

    @property
    def pre_open_interest(self):
        """

        :return:
        """
        return self._pre_open_interest

    @pre_open_interest.setter
    def pre_open_interest(self, value):
        self._pre_open_interest = value

    @property
    def pre_settlement_price(self):
        """

        :return:
        """
        return self._pre_settlement_price

    @pre_settlement_price.setter
    def pre_settlement_price(self, value):
        self._pre_settlement_price = value

    @property
    def change(self):
        """

        :return:
        """
        if self.pre_close_price > 0.0:
            return round((self.last_price / self.pre_close_price - 1.0) * 100.0, 4)
        if self.pre_settlement_price > 0.0:
            return round((self.last_price / self.pre_settlement_price - 1.0) * 100.0, 4)
        return 0.0

    @property
    def if_up_limit(self):
        """

        :return:
        """
        return (self.ask_price1 <= 0.0) & (self.bid_price1 > 0.0)

    @property
    def IfDropLimit(self):
        """

        :return:
        """
        return (self.bid_price1 <= 0.0) & (self.ask_price1 > 0.0)

    @property
    def status(self):
        """

        :return:
        """
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

    @property
    def bid_num_order1(self):
        """

        :return:
        """
        return self.quote.bid_num_order1

    @property
    def ask_num_order1(self):
        """

        :return:
        """
        return self.quote.ask_num_order1

    def to_string(self):
        """
        转换为string
        :return:
        """
        string_builder = ''
        string_builder += str(self.market)
        string_builder += ("[" + self.instrument_id + "]")
        string_builder += ("交易日" + self.trading_date.strftime('%Y/%m/%d')) + ","
        string_builder += ("自然时" + self.date_time.strftime('%Y/%m/%d %H:%M:%S.%f')) + ","
        string_builder += ("最新=" + str(self.last_price)) + ","
        string_builder += ("开=" + str(self.open_price)) + ","
        string_builder += ("高=" + str(self.high_price)) + ","
        string_builder += ("低=" + str(self.low_price)) + ","
        string_builder += ("昨收=" + str(self.pre_close_price)) + ","
        if self.change >= 0.0:
            string_builder += ("涨幅=" + str(round(self.change, 2))) + "%,"
        else:
            string_builder += ("跌幅=" + str(round(self.change, 2))) + "%,"
        string_builder += ("量" + str(self.volume)) + ","
        string_builder += ("额" + str(self.turnover)) + ","
        string_builder += ("[" + str(self.quote.bid_volume1) + "|" + str(self.quote.bid_price1) + "][" + str(self.quote.ask_price1) + "|" +
                           str(self.quote.ask_volume1) + "]")
        return string_builder

    def clone(self):
        """
        克隆
        :return:
        """
        tick = Tick()
        tick._market = self._market
        tick._natural_time = self._natural_time
        tick._trading_date = self._trading_date
        tick._drop_limit_price = self._drop_limit_price
        tick._exchangeId = self._exchange_id
        tick._highPrice = self._high_price
        tick._instrument_id = self._instrument_id
        tick._last_price = self._last_price
        tick._low_price = self._low_price
        tick._open_interest = self._open_interest
        tick._open_price = self._open_price
        tick._pre_close_price = self._pre_close_price
        tick._pre_open_interest = self._pre_open_interest
        tick._pre_settlement_price = self._pre_settlement_price

        if self._quote is not None:
            tick._quote = self._quote.clone()

        tick._turnover = self._turnover
        tick._up_limit_price = self._up_limit_price
        tick._volume = self._volume

        return tick



from enum import Enum, unique


@unique
class EnumOrderSubmitStatus(Enum):
    insert_submitted = '0'
    cancel_submitted = '1'
    modify_submitted = '2'
    accepted = '3'
    insert_rejected = '4'
    cancel_rejected = '5'
    modify_rejected = '6'
    un_known = '-1'



import glob
import os
import shutil
import threading
import zipfile

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import ticker


class FileToolBox:
    @staticmethod
    def save_csv(record_path, df_data, index=False):
        """
        保存DataFrame对象到指定地址
        :param index:
        :param df_data:
        :param record_path:
        """
        try:
            record_deep_dir = os.path.dirname(record_path)
            if not os.path.exists(record_deep_dir):
                os.makedirs(record_deep_dir)

            if os.path.exists(record_path):
                os.remove(record_path)

            print('save_csv:{0}'.format(record_path))
            df_data.to_csv(record_path, encoding='utf-8', index=index)
        except Exception as e:
            print(str(e))

    @staticmethod
    def save_excel(record_path, **kwargs):
        """
        保存Excel
        :param record_path:
        :param kwargs:
        """
        try:
            record_deep_dir = os.path.dirname(record_path)
            if not os.path.exists(record_deep_dir):
                os.makedirs(record_deep_dir)

            if os.path.exists(record_path):
                os.remove(record_path)

            writer = pd.ExcelWriter(record_path)
            for key in kwargs.keys():
                sheet_name = key
                df_data = kwargs[key]
                df_data.to_excel(excel_writer=writer, sheet_name=sheet_name, index=False)
            writer.save()
            writer.close()
            print('save_excel:{0}'.format(record_path))
        except Exception as e:
            print(str(e))

    @staticmethod
    def append_excel(record_path, **kwargs):
        """
        保存Excel
        :param record_path:
        :param kwargs:
        """
        try:
            record_deep_dir = os.path.dirname(record_path)
            if not os.path.exists(record_deep_dir):
                os.makedirs(record_deep_dir)

            if os.path.exists(record_path):
                writer = pd.ExcelWriter(record_path, mode='a')
            else:
                writer = pd.ExcelWriter(record_path)

            for key in kwargs.keys():
                sheet_name = key
                df_data = kwargs[key]
                df_data.to_excel(excel_writer=writer, sheet_name=sheet_name, index=False)
            writer.save()
            writer.close()
            print('append_excel:{0}'.format(record_path))
        except Exception as e:
            print(str(e))

    @staticmethod
    def save_fig(record_path, df_data, x_column, y_columns):
        """
        保存图片
        :param record_path:保存图片路径
        :param df_data:保存的图片的数据
        :param x_column:X轴的列
        :param y_columns:Y轴的列合集
        :return:
        """
        try:
            x_date_key_list = ['datetime64[ns]']
            x_data = df_data[x_column]
            fig = plt.figure(figsize=(15, 8))
            ax = fig.add_subplot(1, 1, 1)

            if x_data.dtypes.name in x_date_key_list:
                N = len(x_data)
                ind = np.arange(N)  # the evenly spaced plot indices

                def format_date(x, pos=None):  # 保证下标不越界,很重要,越界会导致最终plot坐标轴label无显示
                    this_ind = np.clip(int(x + 0.5), 0, N - 1)
                    return df_data[x_column].iat[this_ind]

                for column in y_columns:
                    ax.plot(ind, df_data[column])
                ax.xaxis.set_major_formatter(ticker.FuncFormatter(format_date))
            else:
                for column in y_columns:
                    ax.plot(x_data, df_data[column])

            plt.legend()

            title_name = os.path.basename(record_path)
            plt.title(title_name)

            record_deep_dir = os.path.dirname(record_path)
            if not os.path.exists(record_deep_dir):
                os.makedirs(record_deep_dir)

            if os.path.exists(record_path):
                os.remove(record_path)

            plt.savefig(record_path)
            plt.close(fig)
            print('save_fig:{0}'.format(record_path))
        except Exception as e:
            print(str(e))

    @staticmethod
    def save_twins_fig(record_path, df_data, x_column, main_y_column, secondary_y_column):
        """
        双Y轴绘图保存
        :param record_path:
        :param df_data:
        :param x_column:
        :param main_y_column:
        :param secondary_y_column:
        :return:
        """
        try:
            x_date_key_list = ['datetime64[ns]']
            x_data = df_data[x_column]
            fig = plt.figure(figsize=(15, 8))
            ax = fig.add_subplot(1, 1, 1)

            if x_data.dtypes.name in x_date_key_list:
                N = len(x_data)
                ind = np.arange(N)  # the evenly spaced plot indices

                def format_date(x, pos=None):  # 保证下标不越界,很重要,越界会导致最终plot坐标轴label无显示
                    this_ind = np.clip(int(x + 0.5), 0, N - 1)
                    return df_data[x_column].iat[this_ind]

                ax.plot(ind, df_data[main_y_column])
                ax.xaxis.set_major_formatter(ticker.FuncFormatter(format_date))
                ax2 = ax.twinx()  # 用次y轴画图
                ax2.plot(ind, df_data[secondary_y_column])
            else:
                ax.plot(x_data, df_data[main_y_column])
                ax2 = ax.twinx()  # 用次y轴画图
                ax2.plot(x_data, df_data[secondary_y_column])

            plt.legend()

            title_name = os.path.basename(record_path)
            plt.title(title_name)

            record_deep_dir = os.path.dirname(record_path)
            if not os.path.exists(record_deep_dir):
                os.makedirs(record_deep_dir)

            if os.path.exists(record_path):
                os.remove(record_path)

            plt.savefig(record_path)
            plt.close(fig)
            print('save_fig:{0}'.format(record_path))
        except Exception as e:
            print(str(e))

    @staticmethod
    def save_txt(record_path, msg):
        """
        保存txt
        :param record_path:
        :param msg:
        """
        record_deep_dir = os.path.dirname(record_path)
        if not os.path.exists(record_deep_dir):
            os.makedirs(record_deep_dir)

        if os.path.exists(record_path):
            os.remove(record_path)

        with open(record_path, 'w+') as f:
            f.write(msg)

    @staticmethod
    def recursion_copy(root_path, destination_path, include_dir=None, exclude_dir=None, suffix=None):
        """
        获取py文件的路径
        :param suffix:
        :param include_dir:
        :param include_dir:
        :param root_path: 根路径
        :param destination_path: 目标路径
        :param exclude_dir: 排除文件（当前文件不是目标文件）
        :return: py文件的迭代器
        """
        # 返回所有文件夹绝对路径
        try:
            if exclude_dir is None:
                exclude_dir = []
            # 返回指定的文件夹包含的文件或文件夹的名字的列表
            for file_name in os.listdir(root_path):
                if 'MANIFEST.in' in file_name:
                    continue
                file_path = os.path.join(root_path, file_name)
                if os.path.isdir(file_path) and file_name not in exclude_dir:
                    if include_dir is not None and file_name in include_dir:
                        # 循环遍历文件夹
                        if not os.path.exists(os.path.join(destination_path, file_name)):
                            os.makedirs(os.path.join(destination_path, file_name))
                        FileToolBox.recursion_copy(os.path.join(root_path, file_name), os.path.join(destination_path, file_name), include_dir, exclude_dir, suffix)
                elif os.path.isfile(file_path):
                    # 文件不是排除文件  且不是'.pyc' '.pyx'文件
                    if suffix is not None and os.path.splitext(file_name)[1] in suffix:
                        shutil.copy(os.path.join(root_path, file_name), os.path.join(destination_path, file_name))
                    else:
                        shutil.copy(os.path.join(root_path, file_name), os.path.join(destination_path, file_name))
                else:
                    pass
        except Exception as e:
            print(str(e))

    @staticmethod
    def zip_dir(dir_path, zip_path):
        zip_file = zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED)
        print('[{0}]:压缩:{1}'.format(threading.currentThread().ident, dir_path))
        for root, dir_names, file_names in os.walk(dir_path):
            file_path = root.replace(dir_path, '')  # 去掉根路径，只对目标文件夹下的文件及文件夹进行压缩
            # 循环出一个个文件名
            for file_name in file_names:
                zip_file.write(os.path.join(root, file_name), os.path.join(file_path, file_name))
        zip_file.close()

    @staticmethod
    def unzip_file(dir_path, unzip_file_path):
        # 找到压缩文件夹
        dir_list = glob.glob(dir_path)
        if dir_list:
            # 循环zip文件夹
            for dir_zip in dir_list:
                # 以读的方式打开
                with zipfile.ZipFile(dir_zip, 'r') as f:
                    for file in f.namelist():
                        f.extract(file, path=unzip_file_path)
                os.remove(dir_zip)
# lst_data = [[datetime.datetime(2019, 1, 1), '1', '2', '3'],
#             [datetime.datetime(2019, 1, 2), '2', '3', '4'],
#             [datetime.datetime(2019, 1, 3), '3', '4', '4'],
#             [datetime.datetime(2019, 1, 4), '4', '3', '4'],
#             [datetime.datetime(2019, 1, 5), '5', '5', '2'],
#             [datetime.datetime(2019, 1, 6), '6', '2', '2'],
#             [datetime.datetime(2019, 1, 7), '7', '8', '2'],
#             [datetime.datetime(2019, 1, 8), '8', '9', '5'],
#             [datetime.datetime(2019, 1, 9), '9', '4', '5'],
#             [datetime.datetime(2019, 1, 10), '10', '2', '4'],
#             [datetime.datetime(2019, 1, 11), '11', '8', '7'],
#             [datetime.datetime(2019, 1, 12), '12', '2', '7'],
#             [datetime.datetime(2019, 1, 13), '13', '1', '7'],
#             [datetime.datetime(2019, 1, 14), '14', '0', '4'],]
# df_data = pd.DataFrame(lst_data, columns=['date', 'value1', 'value2', 'value3'])
# record_path = 'D:/a.png'
# FileToolBox.save_fig(record_path, df_data, 'value1', ['value2', 'value3'])



from SapphireQuant.Config.Ctp.EnumServerType import EnumServerType


class CtpCounter:
    """
    CTP柜台信息
    """

    def __init__(self):
        self.quote_servers = []
        self.trade_servers = []

    def add(self, server):
        """
        添加服务器
        :param server:
        """
        if server.server_type == EnumServerType.quote:
            if self.quote_servers is None:
                self.quote_servers = [server]
            else:
                self.quote_servers.append(server)

        if server.server_type == EnumServerType.trade:
            if self.trade_servers is None:
                self.trade_servers = [server]
            else:
                self.trade_servers.append(server)

    def get_servers(self, server_type):
        """
        获取服务器
        :param server_type:
        :return:
        """
        if server_type == EnumServerType.quote:
            return self.quote_servers
        if server_type == EnumServerType.trade:
            return self.trade_servers
        return None



from enum import Enum, unique


@unique
class EnumOrderStatus(Enum):
    全部成交 = '0'
    部分成交 = '1'  # 部分成交还在队列中
    正撤 = '2'  # 部分成交不在队列中 ,    # 未成交不在队列中 = '4'
    已报 = '3'  # 未成交还在队列中
    已撤 = '5'  # 撤单
    未知 = 'a'  # 未知    尚未触发 = 'b'     已触发 = 'c'
    废单 = 'e'  # 废单
    未报 = 'f'

    def chinese_name(self):
        """

        :return:
        """
        if self == EnumOrderStatus.全部成交:
            return '全部成交'
        elif self == EnumOrderStatus.部分成交:
            return '部分成交'
        elif self == EnumOrderStatus.正撤:
            return '正撤'
        elif self == EnumOrderStatus.已报:
            return '已报'
        elif self == EnumOrderStatus.已撤:
            return '已撤'
        elif self == EnumOrderStatus.未知:
            return '未知'
        elif self == EnumOrderStatus.废单:
            return '废单'
        elif self == EnumOrderStatus.未报:
            return '未报'

    def can_cancel(self):
        """
        是否可以撤单
        :return:
        """
        return self == EnumOrderStatus.未报 or self == EnumOrderStatus.已报 or self == EnumOrderStatus.部分成交



import enum
from enum import Enum, unique


@unique
class EnumBarType(enum.IntEnum):
    tick = 0
    second = 1
    minute = 2
    hour = 3
    day = 4
    week = 5
    month = 6
    year = 7
    night_am_pm = 8


from SapphireQuant.Core.QiList import QiList


class TickSeries(QiList):
    """
    BarSeries
    """

    def __init__(self):
        self._instrumentId = ''
        super().__init__(self)

    def get_item_by_date_time(self, date_time):
        """
        根据时间获取元素
        :param date_time:
        :return:
        """
        index = self.get_index(date_time)
        if index > -1:
            return self[index]

        return None

    def contains(self, date_time):
        """
        是否包含指定时间的Bar
        :param date_time:
        :return:
        """
        return self.first is not None and self.last is not None and self.first.date_time <= date_time <= self.last.date_time

    def clone(self):
        """
        克隆
        :return:
        """
        tick_series = TickSeries()
        for i in range(0, self.length):
            tick_series.append(self[i].clone())
        return tick_series

    def get_date_time(self, index):
        """
        获取指定index对应的Bar的begin_time
        :param index:
        :return:
        """
        if self.length > index:
            return self[index].date_time
        return None

    def get_index(self, date_time):
        """
        获取指定date_time所在的index
        :param date_time:
        :return:
        """

        if self.contains(date_time):
            if self.length == 1:
                if abs((self[0].date_time - date_time).total_seconds() * 1000) < 250.0:
                    return 0
            else:
                if self.length >= 2:
                    i = self.length - 1
                    while i > 0:
                        tick = self[i]
                        tick2 = self[i - 1]
                        if tick is not None and tick2 is not None and tick2.date_time <= date_time <= tick.date_time:
                            num = abs((date_time - tick2.date_time).total_seconds() * 1000)
                            num2 = abs((date_time - tick.date_time).total_seconds() * 1000)
                            if num2 <= num:
                                return i
                            return i - 1
                        else:
                            i = i - 1

                pass
        return 0

    def get_ago_index(self, date_time):
        """

        :param date_time:
        :return:
        """
        index = self.get_index(date_time)
        if index > -1:
            return self.length - 1 - index
        return -1

# Test Code
# begin_time = datetime.datetime(2019, 12, 1, 14, 0, 0)
# end_time = datetime.datetime(2019, 12, 1, 14, 0, 3)
# time_delta = begin_time - end_time
# print(time_delta.total_seconds())



from enum import Enum, unique


@unique
class EnumOpenClose(Enum):
    开仓 = 0
    平仓 = 1
    强平 = 2
    平今 = 3
    平昨 = 4
    强减 = 5
    本地强平 = 6

    def chinese_name(self):
        """
        中文名称
        :return:
        """
        if self == EnumOpenClose.开仓:
            return '开仓'
        elif self == EnumOpenClose.平仓:
            return '平仓'
        elif self == EnumOpenClose.强平:
            return '强平'
        elif self == EnumOpenClose.平今:
            return '平今'
        elif self == EnumOpenClose.平昨:
            return '平昨'
        elif self == EnumOpenClose.强减:
            return '强减'
        else:
            return '本地强平'


from SapphireQuant.Config.Ctp.EnumNetworkType import EnumNetworkType


class CtpServer:
    """
    Ctp服务器信息
    """

    def __init__(self, address, group_name, server_type, network_type=EnumNetworkType.telecom):
        self.address = address
        self.group_name = group_name
        self.server_type = server_type
        self.network_type = network_type



from .BaseMatchMachine import BaseMatchMachine
from .LastPriceMatchMachine import LastPriceMatchMachine
from .MatchCenter import MatchCenter
from .OrderPriceMatchMachine import OrderPriceMatchMachine

__all__ = [
    'BaseMatchMachine',
    'LastPriceMatchMachine',
    'MatchCenter',
    'OrderPriceMatchMachine',
]



from .FutureExchangeTradingFrame import FutureExchangeTradingFrame
from .FutureProductTradingFrame import FutureProductTradingFrame
from .FutureTradingFrame import FutureTradingFrame
from .FutureTradingTime import FutureTradingTime
from .LimitValue import LimitValue
from .TradingFrameHelper import TradingFrameHelper
from .TradingFrameManager import TradingFrameManager


__all__ = [
    'FutureExchangeTradingFrame',
    'FutureProductTradingFrame',
    'FutureTradingFrame',
    'FutureTradingTime',
    'LimitValue',
    'TradingFrameHelper',
    'TradingFrameManager',
    'TradingFrameNode',
]



from enum import Enum, unique


@unique
class EnumProductClass(Enum):
    future = 1  # 期货
    option = 2  # 期货期权
    combination = 3  # 组合
    spot = 4  # 即期
    efp = 5  # 期转现
    spot_option = 6  # 现货期权



import datetime
import os
import sys
from decimal import Decimal

from SapphireQuant.Config.TradingDay.TradingDayHelper import TradingDayHelper
from SapphireQuant.Core.CSharpUtils import CSharpUtils
from SapphireQuant.Core.Quote import Quote
from SapphireQuant.Core.Tick import Tick
from SapphireQuant.DataProcessing.BinaryReader import BinaryReader
from SapphireQuant.DataProcessing.BinaryWriter import BinaryWriter


class BinaryTickStream:
    """
    Tick读写流
    """
    CFileFlag = 1262700884  # ('K' << 24) + ('C' << 16) + ('I' << 8) + 'T'
    CFileHeaderLen = 44
    CNaturalDayLen = 8 * 4  # 正常可以放置4个自然日索引
    CTickOffset = CFileHeaderLen + CNaturalDayLen  # 76  Tick数据开始的位置

    # VersionNew
    CTickHeaderLenV0 = 32  # 不带LocalTime
    CTickHeaderLenV1 = 40  # 带LocalTime
    CTickQuoteLen = 4 * 4  # 买一价，卖一价，买一量，卖一量

    # VersionOld
    CFileHeaderLenOld = 64
    CReserveOld = 27
    CTickHeaderLenOld = 32

    def __init__(self, market, instrument_id, exchange_id, file_path):
        self._file_path = ""
        self._instrument_id = ""
        self._exchange_id = ""
        self._market = 0
        self._first_read = False
        self._new_version = True

        # Header
        self._version = 1
        self._quote_level = 1
        self._multi_unit = 1000
        self._trading_day = datetime.datetime.max
        self._pre_close_price = 0.0
        self._pre_settlement_price = 0.0
        self._pre_interest = 0.0
        self._up_limit = 0.0
        self._down_limit = 0.0
        self._open_price = 0.0
        self._tick_count = 0
        self._natural_day_count = 1
        self._natural_day_offset = 0
        self._tick_offset = 0

        # NaturalDay
        self._natural_days = []
        self._natural_day_tick_offset = []

        # 18点后
        self._pre_trading_day1 = datetime.datetime.max
        # 9点前
        self._pre_trading_day2 = datetime.datetime.max

        self._file_path = file_path
        self._instrument_id = instrument_id
        self._exchange_id = exchange_id
        self._first_read = True
        self._market = market

    def _read_header(self, reader):
        """
        读取头
        :param reader:
        :return:
        """
        reader.stream.seek(0, 0)
        flag = reader.read_int32()
        if flag != self.CFileFlag:
            return False
        self._natural_days.clear()
        self._natural_day_tick_offset.clear()
        self._natural_day_count = 0

        self._version = reader.read_int16()
        self._quote_level = reader.read_byte()

        tmp = 1
        multi = reader.read_byte()
        for i in range(0, multi):
            tmp = tmp * 10

        self._multi_unit = tmp

        year = reader.read_int16()
        month = reader.read_byte()
        day = reader.read_byte()

        self._trading_day = datetime.datetime(year, month, day)

        self._pre_close_price = reader.read_int32() / self._multi_unit
        self._pre_settlement_price = reader.read_int32() / self._multi_unit
        self._pre_interest = reader.read_int32()
        self._up_limit = reader.read_int32() / self._multi_unit
        self._down_limit = reader.read_int32() / self._multi_unit
        self._open_price = reader.read_int32() / self._multi_unit
        self._tick_count = reader.read_int32()

        orig = reader.read_int16()
        self._natural_day_count = (orig >> 12)
        self._natural_day_offset = (orig & 0x0fff)
        self._tick_offset = reader.read_int16()
        return True

    def _read_old_header(self, reader):
        """
        读取老版的头
        :param reader:
        """
        reader.stream.seek(0, 0)
        self._natural_days.clear()
        self._natural_day_tick_offset.clear()
        self._natural_day_count = 0

        interval = reader.read_int16()
        bar_type = reader.read_int16()
        if (bar_type & 0xff) != 0:
            raise Exception("错误的tick数据文件标识")

        self._new_version = False
        if ((bar_type >> 10) & 0x3) == 1:
            self._multi_unit = 1000
        else:
            self._multi_unit = 100

        year = reader.read_int16()
        month = reader.read_byte()
        day = reader.read_byte()

        self._trading_day = datetime.datetime(year, month, day)
        self._pre_trading_day1 = TradingDayHelper.get_pre_trading_day(self._trading_day)
        self._pre_trading_day2 = self._pre_trading_day1 + datetime.timedelta(days=1)

        self._pre_close_price = reader.read_int32() / self._multi_unit
        self._pre_settlement_price = reader.read_int32() / self._multi_unit
        self._pre_interest = reader.read_int32()
        self._up_limit = reader.read_int32() / self._multi_unit
        self._down_limit = reader.read_int32() / self._multi_unit
        self._open_price = reader.read_int32() / self._multi_unit
        self._tick_count = reader.read_int32()
        self._quote_level = reader.read_byte()

        reader.read_bytes(self.CReserveOld)

    def _read_natural_days(self, reader):
        """
        读取自然日索引
        :param reader:
        """
        reader.stream.seek(self._natural_day_offset, 0)
        for i in range(0, self._natural_day_count):
            year = reader.read_int16()
            month = reader.read_byte()
            day = reader.read_byte()
            natural_day_tick_offset = reader.read_int32()

            natural_day = datetime.datetime(year, month, day)

            self._natural_days.append(natural_day)
            self._natural_day_tick_offset.append(natural_day_tick_offset)

    def _read_tick(self, reader, natural_day):
        """
        读取Tick
        :param reader:
        :param natural_day:
        """
        tick = Tick()
        tick.market = self._market
        tick.open_price = self._open_price
        tick.pre_close_price = self._pre_close_price
        tick.instrument_id = self._instrument_id
        tick.exchange_id = self._exchange_id
        tick.pre_open_interest = self._pre_interest
        tick.pre_settlement_price = self._pre_settlement_price
        tick.up_limit = self._up_limit
        tick.drop_limit = self._down_limit

        hour = reader.read_byte()
        minute = reader.read_byte()
        second = reader.read_byte()
        milliseconds = reader.read_byte()
        milliseconds *= 10

        tick.date_time = datetime.datetime(natural_day.year, natural_day.month, natural_day.day, hour, minute, second, milliseconds * 1000)

        if self._version == 1:
            c_sharp_ticks = reader.read_int64()
            tick.local_time = CSharpUtils.convert_c_sharp_ticks_to_py_date_time(c_sharp_ticks)
        else:
            tick.local_time = datetime.datetime.today()
        tick.trading_date = self._trading_day
        tick.last_price = reader.read_int32() / self._multi_unit
        tick.high_price = reader.read_int32() / self._multi_unit
        tick.low_price = reader.read_int32() / self._multi_unit
        tick.open_interest = reader.read_int32()
        tick.volume = reader.read_int32()
        tick.turnover = reader.read_double()
        quote = Quote()
        quote.ask_volume1 = reader.read_int32()
        quote.bid_volume1 = reader.read_int32()
        quote.ask_price1 = reader.read_int32() / self._multi_unit
        quote.bid_price1 = reader.read_int32() / self._multi_unit
        tick.quote = quote
        return tick

    def _read_tick_old(self, reader):
        """
        读取Tick
        :param reader:
        """
        tick = Tick()
        tick.market = self._market
        tick.open_price = self._open_price
        tick.pre_close_price = self._pre_close_price
        tick.instrument_id = self._instrument_id
        tick.exchange_id = self._exchange_id
        tick.pre_open_interest = self._pre_interest
        tick.pre_settlement_price = self._pre_settlement_price
        tick.up_limit = self._up_limit
        tick.drop_limit = self._down_limit

        hour = reader.read_byte()
        minute = reader.read_byte()
        second = reader.read_byte()
        milliseconds = reader.read_byte()
        milliseconds *= 10

        tick.trading_date = self._trading_day
        if hour < 7:
            tick.date_time = datetime.datetime(self._pre_trading_day2.year, self._pre_trading_day2.month,
                                               self._pre_trading_day2.day, hour, minute, second, milliseconds)
        elif hour < 18:
            tick.date_time = datetime.datetime(self._trading_day.year, self._trading_day.month,
                                               self._trading_day.day, hour, minute, second, milliseconds)
        else:
            tick.date_time = datetime.datetime(self._pre_trading_day1.year, self._pre_trading_day1.month,
                                               self._pre_trading_day1.day, hour, minute, second, milliseconds)

        tick.last_price = reader.read_int32() / self._multi_unit
        tick.high_price = reader.read_int32() / self._multi_unit
        tick.low_price = reader.read_int32() / self._multi_unit
        tick.open_interest = reader.read_int32()
        tick.volume = reader.read_int32()
        tick.turnover = reader.read_double()
        quote = Quote()
        quote.ask_volume1 = reader.read_int32()
        quote.bid_volume1 = reader.read_int32()
        quote.ask_price1 = reader.read_int32() / self._multi_unit
        quote.bid_price1 = reader.read_int32() / self._multi_unit
        tick.quote = quote
        return tick

    def _read_ticks_by_count(self, tick_series, reader, offset, count):
        """
        读取V1版本的Level1行情
        :param tick_series:
        :param reader:
        :param offset:
        :param count:
        :return:
        """
        pos = self._tick_offset + offset * (self.CTickHeaderLenV1 + self._quote_level * self.CTickQuoteLen)
        reader.stream.seek(pos)
        natural_index = 0
        for natural_index in range(0, len(self._natural_day_tick_offset)):
            if offset < self._natural_day_tick_offset[natural_index]:
                break
            natural_index = natural_index + 1

        natural_index = natural_index - 1
        if natural_index < 0:
            return

        natural_day = self._natural_days[natural_index]
        next_natural_day_tick_offset = 100000000
        if natural_index < len(self._natural_day_tick_offset) - 1:
            next_natural_day_tick_offset = self._natural_day_tick_offset[natural_index + 1] - 1

        read_tick_count = self._tick_count - offset
        if read_tick_count < count:
            read_tick_count = read_tick_count
        else:
            read_tick_count = count

        for i in range(0, read_tick_count):
            tick = self._read_tick(reader, natural_day)
            tick_series.append(tick)

            if i + offset >= next_natural_day_tick_offset:
                natural_index = natural_index + 1
                if natural_index < len(self._natural_days):
                    natural_day = self._natural_days[natural_index]

                next_natural_day_tick_offset = 10000000
                if natural_index < len(self._natural_day_tick_offset) - 1:
                    next_natural_day_tick_offset = self._natural_day_tick_offset[natural_index + 1] - 1

    def _read_ticks_by_count_old(self, tick_series, reader, offset, count):
        """

        :param tick_series:
        :param reader:
        :param offset:
        :param count:
        """
        pos = self.CFileHeaderLenOld + offset * (self.CTickHeaderLenOld + self._quote_level * self.CTickQuoteLen)
        reader.stream.seek(pos)
        read_tick_count = self._tick_count - offset
        if read_tick_count < count:
            read_tick_count = read_tick_count
        else:
            read_tick_count = count

        for i in range(0, read_tick_count):
            tick = self._read_tick_old(reader)
            tick_series.append(tick)

    def _read_ticks_by_time(self, tick_series, reader, begin_time, end_time):
        """

        :param tick_series:
        :param reader:
        :param begin_time:
        :param end_time:
        :return:
        """
        reader.stream.seek(self._tick_offset, 0)
        if len(self._natural_days) == 0:
            return

        natural_index = 0
        natural_day = self._natural_days[natural_index]
        next_natural_day_tick_offset = sys.maxsize
        if natural_index < len(self._natural_day_tick_offset) - 1:
            next_natural_day_tick_offset = self._natural_day_tick_offset[natural_index + 1] - 1

        blk_len = self.CTickHeaderLenV1 + self._quote_level * self.CTickQuoteLen - 4

        for i in range(self._tick_count):
            hour = reader.read_byte()
            minute = reader.read_byte()
            second = reader.read_byte()
            milliseconds = reader.read_byte()
            milliseconds *= 10

            date_time = datetime.datetime(natural_day.year, natural_day.month, natural_day.day, hour, minute, second, milliseconds)

            if date_time > end_time:
                return

            if date_time < begin_time:
                if i >= next_natural_day_tick_offset:
                    natural_index += 1

                    if natural_index < len(self._natural_days):
                        natural_day = self._natural_days[natural_index]

                    next_natural_day_tick_offset = sys.maxsize
                    if natural_index < (len(self._natural_day_tick_offset) - 1):
                        next_natural_day_tick_offset = self._natural_day_tick_offset[natural_index + 1] - 1
                reader.read_bytes(blk_len)
                continue

            tick = self._read_tick(reader, natural_day)
            tick_series.append(tick)

            if i >= next_natural_day_tick_offset:
                natural_index += 1
                if natural_index < len(self._natural_days):
                    natural_day = self._natural_days[natural_index]

                next_natural_day_tick_offset = sys.maxsize
                if natural_index < len(self._natural_day_tick_offset) - 1:
                    next_natural_day_tick_offset = self._natural_day_tick_offset[natural_index + 1] - 1

    def _read_ticks_by_time_old(self, tick_series, reader, begin_time, end_time):
        """

        :param tick_series:
        :param reader:
        :param begin_time:
        :param end_time:
        :return:
        """
        reader.stream.seek(self.CFileHeaderLenOld, 0)
        blk_len = self.CTickHeaderLenV0 + self._quote_level * self.CTickQuoteLen - 4

        for i in range(self._tick_count):
            hour = reader.read_byte()
            minute = reader.read_byte()
            second = reader.read_byte()
            millisecond = reader.read_byte()
            millisecond *= 10

            if hour < 7:
                tick_time = datetime.datetime(self._pre_trading_day2.year, self._pre_trading_day2.month,
                                              self._pre_trading_day2.day, hour, minute, second, millisecond)
            elif hour < 18:
                tick_time = datetime.datetime(self._trading_day.year, self._trading_day.month,
                                              self._trading_day.day, hour, minute, second, millisecond)
            else:
                tick_time = datetime.datetime(self._pre_trading_day1.year, self._pre_trading_day1.month,
                                              self._pre_trading_day1.day, hour, minute, second, millisecond)

            if tick_time > end_time:
                return

            if tick_time < begin_time:
                reader.read_bytes(blk_len)
                continue

            tick = self._read_tick_old(reader)
            tick_series.append(tick)

    def read_last_tick(self):
        """
        根据个数读取
        :return:
        """
        if not os.path.exists(self._file_path):
            _firstRead = True
            print("Read Future Tick Data Failed:{},File Do Not Exist....".format(self._file_path))
            return None

        stream = open(self._file_path, 'rb')
        reader = BinaryReader(stream)

        if self._first_read:
            if self._read_header(reader):
                self._read_natural_days(reader)
            else:
                self._read_old_header(reader)
            _firstRead = False

        tick_series = []
        offset = self._tick_count - 1
        count = 1
        if self._new_version:
            if self._quote_level == 1:
                self._read_ticks_by_count(tick_series, reader, offset, count)
            else:
                raise Exception("不支持" + str(self._quote_level) + "档盘口")
        else:
            if self._quote_level == 1:
                self._read_ticks_by_count_old(tick_series, reader, offset, count)
            else:
                raise Exception("不支持" + str(self._quote_level) + "档盘口")
        stream.close()
        if len(tick_series) > 0:
            return tick_series[0]
        return None

    def read_by_count(self, tick_series, offset, count):
        """
        根据个数读取
        :param tick_series:
        :param offset:跳多少个Tick去取数据
        :param count:
        :return:
        """
        if not os.path.exists(self._file_path):
            _firstRead = True
            print("Read Tick Data Error,Do Not Exist File:{0}".format(self._file_path))
            return False

        stream = open(self._file_path, 'rb')
        reader = BinaryReader(stream)

        if self._first_read:
            if self._read_header(reader):
                self._read_natural_days(reader)
            else:
                self._read_old_header(reader)
            _firstRead = False

        if self._new_version:
            if self._quote_level == 1:
                self._read_ticks_by_count(tick_series, reader, offset, count)
            else:
                raise Exception("Do Not Support {0} Quote Level......".format(str(self._quote_level)))
        else:
            if self._quote_level == 1:
                self._read_ticks_by_count_old(tick_series, reader, offset, count)
            else:
                raise Exception("不支持" + str(self._quote_level) + "档盘口")
        stream.close()
        return True

    def read_by_time(self, tick_series, begin_time, end_time):
        """
        根据时间读取Tick
        :param tick_series:
        :param begin_time:
        :param end_time:
        :return:
        """
        if (begin_time is None) and (end_time is None):
            return self.read_by_count(0, 0, sys.maxsize)

        if end_time is None:
            end_time = datetime.datetime.max

        try:
            if not os.path.exists(self._file_path):
                _firstRead = True
                print("Read Tick Data Error,Do Not Exist File:{0}".format(self._file_path))
                return False

            stream = open(self._file_path, 'rb')
            reader = BinaryReader(stream)

            if self._first_read:
                if self._read_header(reader):
                    self._read_natural_days(reader)
                else:
                    self._read_old_header(reader)
                _firstRead = False

            if self._new_version:
                if self._quote_level == 1:
                    self._read_ticks_by_time(tick_series, reader, begin_time, end_time)
                else:
                    raise Exception("不支持" + str(self._quote_level) + "档盘口")
            else:
                if self._quote_level == 1:
                    self._read_ticks_by_time_old(tick_series, reader, begin_time, end_time)
                else:
                    raise Exception("不支持" + str(self._quote_level) + "档盘口")
            stream.close()
            return True
        except Exception as e:
            print("Read Tick Data By Time Error:{0}".format(str(e)))

        return False

    def _write_header(self, writer):
        """
        写头
        :param writer:
        """
        writer.stream.seek(0, 0)
        writer.write_int32(self.CFileFlag)
        writer.write_int16(self._version)
        writer.write_byte(self._quote_level)
        multi = 0
        tmp = self._multi_unit
        while tmp != 0:
            tmp = int(tmp / 10)
            multi += 1

        multi -= 1
        if multi < 0:
            raise Exception("write multi error")

        writer.write_byte(multi)
        writer.write_int16(self._trading_day.year)
        writer.write_byte(self._trading_day.month)
        writer.write_byte(self._trading_day.day)
        self._write_float_with_decimal(writer, self._pre_close_price)
        self._write_float_with_decimal(writer, self._pre_settlement_price)
        self._write_float_with_decimal(writer, self._pre_interest)
        self._write_float_with_decimal(writer, self._up_limit)
        self._write_float_with_decimal(writer, self._down_limit)
        self._write_float_with_decimal(writer, self._open_price)

        buffer = bytearray(self.CNaturalDayLen + 8)
        writer.write_bytes(buffer)

    def _write_natural_days(self, writer):
        """

        :param writer:
        """
        writer.stream.seek(self.CFileHeaderLen - 8, 0)
        writer.write_int32(self._tick_count)
        tmp = (len(self._natural_days) << 12) + self.CFileHeaderLen
        writer.write_int16(tmp)
        writer.write_int16(self.CTickOffset)
        for i in range(len(self._natural_days)):
            natural_day = self._natural_days[i]
            natural_day_tick_offset = self._natural_day_tick_offset[i]

            writer.write_int16(natural_day.year)
            writer.write_byte(natural_day.month)
            writer.write_byte(natural_day.day)
            writer.write_int32(natural_day_tick_offset)

    def _write_ticks(self, tick_list, writer, offset, count):
        """
        写入
        """
        natural_day = datetime.datetime.max
        if len(self._natural_days) > 0:
            natural_day = self._natural_days[-1]

        end = offset + count
        for i in range(offset, end, 1):
            if i >= end:
                break
            tick = tick_list[i]
            tick_date = datetime.datetime(tick.date_time.year, tick.date_time.month, tick.date_time.day)
            if tick_date != natural_day:
                if len(self._natural_days) > 0:
                    if tick_date < self._natural_days[-1]:
                        print('Tick Date Error,Tick Natural Date={0},File Date={1}'.format(tick_date.strftime('%Y%m%d'), self._natural_days[-1].strftime('%Y%m%d')))
                    if tick_date == self._natural_days[-1]:
                        natural_day = self._natural_days[-1]
                    else:
                        natural_day = tick_date
                        natural_tick_offset = self._tick_count
                        self._natural_days.append(natural_day)
                        self._natural_day_tick_offset.append(natural_tick_offset)
                else:
                    natural_day = tick_date
                    natural_tick_offset = self._tick_count
                    self._natural_days.append(natural_day)
                    self._natural_day_tick_offset.append(natural_tick_offset)

            writer.write_byte(tick.date_time.hour)
            writer.write_byte(tick.date_time.minute)
            writer.write_byte(tick.date_time.second)
            writer.write_byte(int(tick.date_time.microsecond / 1000 / 10))
            c_sharp_ticks = CSharpUtils.get_c_sharp_ticks(tick.local_time)
            writer.write_int64(c_sharp_ticks)

            self._write_float_with_decimal(writer, tick.last_price)
            self._write_float_with_decimal(writer, tick.high_price)
            self._write_float_with_decimal(writer, tick.low_price)
            self._write_float_with_decimal(writer, tick.open_interest)

            writer.write_int32(tick.volume)
            writer.write_double(tick.turnover)
            writer.write_int32(tick.quote.ask_volume1)
            writer.write_int32(tick.quote.bid_volume1)
            self._write_float_with_decimal(writer, tick.ask_price1)
            self._write_float_with_decimal(writer, tick.bid_price1)
            self._tick_count += 1

    def _write_float_with_decimal(self, writer, value):
        """

        :param writer:
        :param value:
        """
        if value < sys.float_info.max:
            value = Decimal.from_float(value)
            writer.write_int32(int(round(value * Decimal.from_float(self._multi_unit), 0)))
        else:
            writer.write_int32(0)

    def _update_tick_open_price(self, tick_list, writer):
        """
        更新开盘价
        :param tick_list:
        :param writer:
        """
        if len(tick_list) > 0:
            open_price = tick_list[0].open_price
            if abs(open_price - self._open_price) >= 0.000000001:
                writer.stream.seek(32, 0)
                self._write_float_with_decimal(writer, open_price)

    def write(self, tick_list, start, length, quote_level):
        """

        :param tick_list:
        :param start:
        :param length:
        :param quote_level:
        :return:
        """
        if len(tick_list) == 0:
            return True

        try:
            self._natural_day_count = 0
            self._natural_days.clear()
            self._natural_day_tick_offset.clear()

            self._multi_unit = 1000

            self._trading_day = tick_list[0].trading_date
            self._pre_close_price = tick_list[0].pre_close_price
            self._open_price = tick_list[0].open_price
            self._pre_interest = tick_list[0].pre_open_interest
            self._pre_settlement_price = tick_list[0].pre_settlement_price
            self._up_limit = tick_list[0].up_limit
            self._down_limit = tick_list[0].drop_limit
            self._quote_level = quote_level

            count = len(tick_list) - start
            if length < count:
                count = length

            self._tick_count = 0
            stream = open(self._file_path, 'wb')
            writer = BinaryWriter(stream)
            self._write_header(writer)

            if self._quote_level == 1:
                self._write_ticks(tick_list, writer, start, count)
            elif self._quote_level == 5:
                pass

            self._write_natural_days(writer)
            stream.flush()
            stream.close()

            return True
        except Exception as e:
            print('Write Ticks Error:{0}'.format(str(e)))

    def append(self, tick_list, start, length, quote_level):
        """
        追加Tick
        :param tick_list:
        :param start:
        :param length:
        :param quote_level:
        :return:
        """
        if not os.path.exists(self._file_path):
            print('Append Tick, File Not Exist, Cover:{0}'.format(self._file_path))
            record_deep_dir = os.path.dirname(self._file_path)
            if not os.path.exists(record_deep_dir):
                os.makedirs(record_deep_dir)
            return self.write(tick_list, start, length, quote_level)
        if os.path.getsize(self._file_path) < self.CFileHeaderLen + self.CNaturalDayLen:
            print('Append Tick, File Size Error, Cover:{0}'.format(self._file_path))
            return self.write(tick_list, start, length, quote_level)
        if (os.path.getsize(self._file_path) - self.CTickOffset) % (self.CTickHeaderLenV1 + quote_level * self.CTickQuoteLen) != 0:
            print('Append Tick, File Layout Error, Cover:{0}'.format(self._file_path))
            return self._write_ticks(tick_list, start, length, quote_level)
        try:
            stream = open(self._file_path, 'rb+')
            reader = BinaryReader(stream)
            if self._first_read:
                self._read_header(reader)
                self._read_natural_days(reader)
                self._first_read = True

            if self._quote_level != quote_level:
                raise Exception("Append Tick Error,Quote Level:File={0},Tick={1}".format(self._quote_level, quote_level))

            tick_count = int((os.path.getsize(self._file_path) - self.CTickOffset) / (self.CTickHeaderLenV1 + quote_level * self.CTickQuoteLen))
            if tick_count != self._tick_count:
                print("File Tick Count:{0} Not Equals Cache Tick Count:{1},File Path:{2}".format(tick_count, self._tick_count, self._file_path))

            self._tick_count = tick_count

            count = len(tick_list) - start
            if length < count:
                count = length

            writer = BinaryWriter(stream)
            self._update_tick_open_price(tick_list, writer)
            stream.seek(0, 2)
            if self._quote_level == 1:
                self._write_ticks(tick_list, writer, start, count)
            else:
                pass

            self._write_natural_days(writer)
            stream.flush()
            stream.close()
            return True

        except Exception as e:
            self._first_read = False
            print("Append Tick Error:{0},{1}".format(self._file_path, str(e)))
        return False



from enum import Enum, unique


@unique
class EnumBuySell(Enum):
    买入 = 0
    卖出 = 1

    def chinese_name(self):
        """

        :return:
        """
        if self == EnumBuySell.买入:
            return '买入'
        else:
            return '卖出'



import datetime
import uuid

from SapphireQuant.Config.TradingDay import TradingDayHelper

from SapphireQuant.Core.Enum import EnumOpenClose, EnumBuySell, EnumPositionDirection, EnumMarket, EnumOrderStatus
from SapphireQuant.Core.Futures import FutureAsset, FuturePosition
from SapphireQuant.DataStorage.Cache import PositionKey, FutureData


class OldFutureCache:
    """
    期货缓存，老版本的期货缓存管理，已经弃用，不带持仓明细，清算不准确
    """
    def __init__(self, transaction_rate, long_margin_rate, short_margin_rate, instrument_manager, data_handler):
        self._trading_date = datetime.datetime.today()
        self._position_map = {}
        self._order_map = {}
        self._trades = []
        self._asset = FutureAsset()

        self._data_map = {}
        self._transaction_rate = transaction_rate
        self._longMargin_rate = long_margin_rate
        self._shortMargin_rate = short_margin_rate
        self.instrument_manager = instrument_manager
        self.data_handler = data_handler

    def on_bar(self, bar):
        """
        on_bar维护缓存
        :param bar:
        """
        if bar.instrument_id not in self._data_map.keys():
            self._data_map[bar.instrument_id] = {}
        self._data_map[bar.instrument_id]['close'] = bar.close
        self._data_map[bar.instrument_id]['pre_close'] = bar.pre_close
        self._data_map[bar.instrument_id]['date_time'] = bar.end_time

    def on_tick(self, tick):
        """
        on_tick维护缓存
        :param tick:
        """
        if tick.instrument_id not in self._data_map.keys():
            self._data_map[tick.instrument_id] = {}
        self._data_map[tick.instrument_id]['close'] = tick.last_price
        self._data_map[tick.instrument_id]['pre_close'] = tick.pre_close_price
        self._data_map[tick.instrument_id]['date_time'] = tick.date_time

    def add_or_update_order(self, order):
        """
        添加或者更新order
        :param order:
        """
        if order.guid in self._order_map.keys():
            self._order_map[order.guid].update(order)
        else:
            self._order_map[order.guid] = order

    def add_trade(self, trade):
        """
        添加trade
        :param trade:
        """
        self._trades.append(trade)

    def calc_margin(self, position, volume_multiple):
        """
        计算保证金
        :param position:
        :param volume_multiple:
        :return:
        """
        if position.direction == EnumPositionDirection.多头:
            return position.volume * position.last_price * volume_multiple * self._longMargin_rate
        return position.volume * position.last_price * volume_multiple * self._shortMargin_rate

    def on_order(self, order):
        """
        on_order触发
        :param order:
        """
        self.add_or_update_order(order)

    def on_trade(self, trade):
        """
        on_trade触发
        :param trade:
        :return:
        """
        future_info = self.instrument_manager[trade.instrument_id]
        multiplier = future_info.volume_multiple
        d_hand_fee = trade.trade_price * trade.trade_volume * multiplier * self._transaction_rate
        old_margin = 0
        # 持仓改变
        if trade.open_close == EnumOpenClose.开仓:
            if trade.direction == EnumBuySell.买入:
                position_direction = EnumPositionDirection.多头
            else:
                position_direction = EnumPositionDirection.空头
            key = PositionKey(trade.product_code, trade.strategy_id, trade.instrument_id, position_direction)
            if key.to_string() in self._position_map.keys():
                position = self._position_map[key.to_string()]

                old_margin = self.calc_margin(position, multiplier)
                position.open_cost = (position.open_cost * position.volume + trade.trade_price * trade.trade_volume) / (position.volume + trade.trade_volume)
                position.volume += trade.trade_volume
                position.td_volume += trade.trade_volume
                position.td_open_trade += trade.trade_volume
                position.last_price = trade.trade_price
                position.transaction_cost += d_hand_fee
            else:

                # 添加持仓
                position = FuturePosition()
                position.product_code = trade.product_code
                position.strategy_id = trade.strategy_id
                # position.ProjectId = trade.ProjectId
                position.trading_date = trade.trading_date
                position.exchange_id = trade.exchange_id
                position.instrument_id = trade.instrument_id
                position.instrument_name = trade.instrument_name
                position.open_cost = trade.trade_price
                position.volume = trade.trade_volume
                position.td_volume = trade.trade_volume
                position.transaction_cost = d_hand_fee
                position.direction = position_direction
                position.td_open_trade = trade.trade_volume

                self._position_map[key.to_string()] = position
        else:  # 平仓
            if trade.direction == EnumBuySell.买入:
                position_direction = EnumPositionDirection.空头
            else:
                position_direction = EnumPositionDirection.多头

            # // 反向持仓
            key = PositionKey(trade.product_code, trade.strategy_id, trade.instrument_id, position_direction)
            position = None
            if key.to_string() in self._position_map.keys():
                position = self._position_map[key.to_string()]
            if position is None:
                return

            old_margin = self.calc_margin(position, multiplier)

            pre_close = self.data_handler.load_pre_close_by_tick(EnumMarket.期货, position.instrument_id, self._trading_date)

            d_close_profit = 0
            position.volume -= trade.trade_volume
            if trade.trade_volume >= position.yd_volume:
                close_yd_volume = position.yd_volume
            else:
                close_yd_volume = trade.trade_volume

            close_td_volume = trade.trade_volume - close_yd_volume
            # 平今仓
            if trade.direction == EnumBuySell.买入:
                d_close_profit += close_td_volume * (trade.trade_price - position.open_cost) * multiplier * -1  # 平仓是有盈亏的
            else:
                d_close_profit += close_td_volume * (trade.trade_price - position.open_cost) * multiplier * 1  # 平仓是有盈亏的
            position.td_volume -= close_td_volume
            # 平老仓
            if trade.direction == EnumBuySell.买入:
                d_close_profit += close_yd_volume * (trade.trade_price - pre_close) * multiplier * -1  # 平仓是有盈亏的
            else:
                d_close_profit += close_yd_volume * (trade.trade_price - pre_close) * multiplier * 1  # 平仓是有盈亏的
            position.yd_volume -= close_yd_volume

            position.close_profit += d_close_profit
            position.td_close_trade += trade.trade_volume
            position.last_price = trade.trade_price
            position.transaction_cost += d_hand_fee

            self._asset.close_profit += d_close_profit
            self._asset.acc_close_profit += d_close_profit
            self._asset.cash += d_close_profit
            self._asset.fund_balance += d_close_profit

        self._asset.cash -= d_hand_fee
        self._asset.fund_balance -= d_hand_fee
        self._asset.fee += d_hand_fee

        # 统一根据持仓价格计算保证金

        margin = self.calc_margin(position, multiplier)
        self._asset.margin += margin - old_margin
        self._asset.available_cash = self._asset.fund_balance - self._asset.margin

        self.add_trade(trade)

    def day_begin(self, trading_date, future_data):
        """
        当天开始，缓存初始化
        :param trading_date:
        :param future_data:
        """
        self._trading_date = trading_date
        self._position_map.clear()
        self._order_map.clear()
        self._trades.clear()
        self._asset = future_data.next_asset
        for position in future_data.next_positions:
            key = PositionKey(position.product_code, position.strategy_id, position.instrument_id, position.direction)
            self._position_map[key.to_string()] = position

    def day_end(self):
        """
        当天结束，缓存清算
        :return:
        """
        next_trading_day = TradingDayHelper.get_next_trading_day(self._trading_date)
        self._asset.position_profit = 0
        d_deposit = 0
        d_market_value = 0
        next_positions = []
        for position in self._position_map.values():
            future_info = self.instrument_manager[position.instrument_id]
            d_multiplier = future_info.volume_multiple

            if position.instrument_id in self._data_map.keys():
                close = self._data_map[position.instrument_id]['close']
                pre_close = self._data_map[position.instrument_id]['pre_close']
            else:
                close = self.data_handler.load_day_bar_series(EnumMarket.期货, position.instrument_id, self._trading_date)[-1].close
                pre_close = self.data_handler.load_pre_close_by_tick(EnumMarket.期货, position.instrument_id, self._trading_date)

            position.position_profit = 0.0
            position.last_price = 0
            if close > 0:
                position_profit = 0.0
                # 今仓 持仓盈亏计算
                if position.direction == EnumPositionDirection.多头:
                    position_profit += (close - position.open_cost) * position.td_volume * d_multiplier * 1
                else:
                    position_profit += (close - position.open_cost) * position.td_volume * d_multiplier * -1

                # 历史 持仓盈亏计算
                if position.direction == EnumPositionDirection.多头:
                    position_profit += (close - pre_close) * position.yd_volume * d_multiplier * 1
                else:
                    position_profit += (close - pre_close) * position.yd_volume * d_multiplier * -1

                self._asset.position_profit += position_profit
                self._asset.fund_balance += position_profit
                position.last_price = close
                position.position_profit = position_profit
                if position.direction == EnumPositionDirection.多头:
                    position.market_value = close * position.volume * d_multiplier * 1
                else:
                    position.market_value = close * position.volume * d_multiplier * -1
                d_market_value += position.market_value

            d_deposit += self.calc_margin(position, d_multiplier)

            if position.volume > 0:
                next_position = position.clone()
                next_position.guid = uuid.uuid1()
                next_position.td_volume = 0
                next_position.yd_volume = position.volume
                next_position.close_volume = 0
                next_position.trading_date = next_trading_day
                next_position.td_close_trade = 0
                next_position.td_open_trade = 0
                next_position.close_profit = 0
                next_position.position_profit = 0
                next_position.num_holding_day = position.num_holding_day + 1
                next_positions.append(next_position)

        self._asset.margin = d_deposit
        self._asset.market_value = d_market_value
        self._asset.available_cash = self._asset.fund_balance - self._asset.margin

        next_asset = self._asset.clone()
        next_asset.guid = uuid.uuid1()
        next_asset.trading_date = next_trading_day
        next_asset.close_profit = 0
        next_asset.position_profit = 0
        next_asset.fee = 0
        next_asset.pre_fund_balance = self._asset.fund_balance

        next_future_data = FutureData()
        next_future_data.asset = self._asset
        next_future_data.positions = self._position_map.values()
        next_future_data.orders = self._order_map.values()
        next_future_data.trades = []
        next_future_data.next_asset = next_asset
        next_future_data.next_positions = next_positions

        for trade in self._trades:
            next_future_data.trades.append(trade)

        return next_future_data

    def get_position(self, product_code, strategy_id, instrument_id, direction):
        """
        获取缓存持仓
        :param product_code:
        :param strategy_id:
        :param instrument_id:
        :param direction:
        :return:
        """
        key = PositionKey(product_code, strategy_id, instrument_id, direction)
        key = key.to_string()
        if key in self._position_map.keys():
            return self._position_map[key]
        else:
            # 添加持仓
            position = FuturePosition()
            position.product_code = product_code
            position.strategy_id = strategy_id
            # position.ProjectId = trade.ProjectId
            position.trading_date = self._trading_date
            position.exchange_id = ''
            position.instrument_id = instrument_id
            position.instrument_name = ''
            position.open_cost = 0
            position.volume = 0
            position.td_volume = 0
            position.transaction_cost = 0
            position.direction = direction
            position.td_open_trade = 0

            self._position_map[key] = position
            return position

    def get_future_asset(self):
        """
        获取期货资金明细
        :return:
        """
        return self._asset

    def get_future_trades(self):
        """
        获取期货成交
        :return:
        """
        return self._trades

    def get_future_orders(self):
        """
        获取期货委托
        :return:
        """
        return list(self._order_map.values())

    def get_can_cancel_orders(self):
        """
        获取可撤委托
        :return:
        """
        order_series = []
        for order in self._order_map.values():
            if order.order_status != EnumOrderStatus.全部成交 and order.order_status != EnumOrderStatus.已撤:
                order_series.append(order)

        return order_series

    def get_future_positions(self):
        """
        获取期货持仓
        :return:
        """
        position_series = list(self._position_map.values())
        return position_series

    def get_future_order(self, guid):
        """
        获取期货委托
        :param guid:
        :return:
        """
        if guid in self._order_map.keys():
            return self._order_map[guid]
        return None



from SapphireQuant.Config.TradingFrame.TradingFrameHelper import TradingFrameHelper


class FutureExchangeTradingFrame:
    """
    期货市场交易框架
    """
    def __init__(self):
        self._id = ""
        self._name = ""
        self._trading_time = ""
        self._products = []

    @property
    def id(self):
        """

        :return:
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def name(self):
        """

        :return:
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def trading_time(self):
        """

        :return:
        """
        return self._trading_time

    @trading_time.setter
    def trading_time(self, value):
        self._trading_time = value

    @property
    def products(self):
        """

        :return:
        """
        return self._products

    @products.setter
    def products(self, value):
        self._products = value

    @property
    def trading_time_slices(self):
        """

        :return:
        """
        return TradingFrameHelper.parse_time_slice(self._trading_time)
